# -*- coding: UTF-8 -*-
# Time: 1/6/18
# File: crowdsource_server.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.web
from tornado import gen
import os
import pipes
import json
import uuid
from tornado.options import define, options
from antgo.utils import logger
import multiprocessing
import requests
from antgo import config
import signal
import copy
import shutil
import socket
import random
import sys
from antgo.crowdsource.utils import *
Config = config.AntConfig


class BaseHandler(tornado.web.RequestHandler):
  @property
  def task_name(self):
    return self.settings['task_name']

  @property
  def html_template(self):
    return self.settings['html_template']

  @property
  def keywords_template(self):
    return self.settings.get('keywords_template', {})

  @property
  def db(self):
    return self.settings['db']

  @property
  def port(self):
    return self.settings['port']

  @property
  def token(self):
    return self.settings['token']

  @property
  def request_queue(self):
    return self.settings['request_queue']

  @property
  def response_queue(self):
    return self.settings['response_queue']


class IndexHandler(BaseHandler):
  def get(self):
    session_id = self.get_cookie('sessionid')
    if session_id is None:
      session_id = str(uuid.uuid4())
      self.set_secure_cookie('sessionid', session_id)

    if session_id not in self.db['user']:
      self.db['user'].append(session_id)

    # render page
    self.keywords_template.update({'task': self.task_name})
    self.render(self.html_template, **self.keywords_template)


class HeartBeatHandler(tornado.web.RequestHandler):
  def get(self):
    self.write(json.dumps({'ALIVE': True}))


class ClientQuery(BaseHandler):
  @gen.coroutine
  def post(self):
    if self.get_cookie('sessionid') not in self.db['user']:
      self.set_status(500)
      self.finish()
      return

    client_query = {}
    client_query['QUERY'] = self.get_argument('QUERY')
    client_query['CLIENT_ID'] = self.get_cookie('sessionid')
    client_query['CLIENT_RESPONSE'] = {}
    client_query['CLIENT_RESPONSE']['WORKSITE'] = self.get_argument('CLIENT_RESPONSE_WORKSITE', None)
    client_query['CLIENT_RESPONSE']['CONCLUSION'] = self.get_argument('CLIENT_RESPONSE_CONCLUSION', None)
    client_query['QUERY_INDEX'] = int(self.get_argument('QUERY_INDEX', -1))
    client_query['QUERY_STATUS'] = self.get_argument('QUERY_STATUS', '')

    # 1.step get process result
    self.request_queue.put(client_query)
    server_response = self.response_queue.get()

    # 3.step return server response
    self.write(json.dumps(server_response))


class PrefixRedirectHandler(BaseHandler):
  def get(self):
    static_pi = self.request.uri.find('static')
    path = self.request.uri[static_pi:]
    self.redirect('http://127.0.0.1:%d/%s'%(self.port, path), permanent=False)


class GracefulExitException(Exception):
  @staticmethod
  def sigterm_handler(signum, frame):
    raise GracefulExitException()


def crowdsrouce_server_start(parent_id,
                             experiment_id,
                             app_token,
                             dump_dir,
                             task_name,
                             html_template,
                             keywords_template,
                             server_port,
                             crowdsource_info={},
                             request_queue=None,
                             response_queue=None):
  # register sig
  signal.signal(signal.SIGTERM, GracefulExitException.sigterm_handler)
  
  # log
  logger.info('crowdsource server prepare serving on %d'%server_port)

  # 0.step define tornado http server port
  define('port', default=server_port, help="run on the given port", type=int)

  try:
    # 0.step prepare static resource to dump_dir
    if not os.path.exists(os.path.join(dump_dir, 'static')):
      os.makedirs(os.path.join(dump_dir, 'static'))

    crowdsource_static_folder = os.path.join(dump_dir, 'static')

    static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
    for static_file in os.listdir(os.path.join(static_folder, 'resource', 'static')):
      if static_file[0] == '.':
        continue

      shutil.copy(os.path.join(static_folder, 'resource', 'static', static_file), crowdsource_static_folder)

    # 1.step prepare template resource
    crowdsource_tempate_dir = os.path.join(dump_dir, 'templates')
    if not os.path.exists(crowdsource_tempate_dir):
      os.makedirs(crowdsource_tempate_dir)

    shutil.copy(os.path.join(static_folder, 'resource', 'templates', 'crowdsource.html'), crowdsource_tempate_dir)
    if os.path.exists(os.path.join(static_folder, 'resource', 'templates', html_template)):
      shutil.copy(os.path.join(static_folder, 'resource', 'templates', html_template), crowdsource_tempate_dir)

    if os.path.exists(os.path.join(os.curdir, html_template)):
      shutil.copy(os.path.join(os.curdir, html_template), crowdsource_tempate_dir)

    # 3.step launch crowdsource server (http server)
    tornado.options.parse_command_line()
    settings={'template_path': crowdsource_tempate_dir,
              'static_path': crowdsource_static_folder,
              'task_name': task_name,
              'html_template': html_template,
              'keywords_template': keywords_template,
              'cookie_secret': str(uuid.uuid4()),
              'port': server_port,
              'db': {'user': []},
              'token': app_token,
              'request_queue': request_queue,
              'response_queue': response_queue}
    app = tornado.web.Application(handlers=[(r"/", IndexHandler),
                                            (r"/query", ClientQuery),
                                            (r"/heartbeat", HeartBeatHandler),
                                            (r"/.*/static/.*", PrefixRedirectHandler)],
                                  **settings)
    http_server = tornado.httpserver.HTTPServer(app)

    http_server.listen(options.port)
    tornado.ioloop.IOLoop.instance().start()
  except GracefulExitException:
    logger.info('exit crowdsource server')
  except Exception as err:
    # exception
    logger.error(err.message)
  finally:
    sys.exit(0)



# Copyright (c) OpenMMLab. All rights reserved.
import torch


def cast_tensor_type(x, scale=1., dtype=None):
    if dtype == 'fp16':
        # scale is for preventing overflows
        x = (x / scale).half()
    return x


def fp16_clamp(x, min=None, max=None):
    if not x.is_cuda and x.dtype == torch.float16:
        # clamp for cpu float16, tensor fp16 has no clamp implementation
        return x.float().clamp(min, max).half()

    return x.clamp(min, max)


class BboxOverlaps2D:
    """2D Overlaps (e.g. IoUs, GIoUs) Calculator."""

    def __init__(self, scale=1., dtype=None):
        self.scale = scale
        self.dtype = dtype

    def __call__(self, bboxes1, bboxes2, mode='iou', is_aligned=False):
        """Calculate IoU between 2D bboxes.

        Args:
            bboxes1 (Tensor): bboxes have shape (m, 4) in <x1, y1, x2, y2>
                format, or shape (m, 5) in <x1, y1, x2, y2, score> format.
            bboxes2 (Tensor): bboxes have shape (m, 4) in <x1, y1, x2, y2>
                format, shape (m, 5) in <x1, y1, x2, y2, score> format, or be
                empty. If ``is_aligned `` is ``True``, then m and n must be
                equal.
            mode (str): "iou" (intersection over union), "iof" (intersection
                over foreground), or "giou" (generalized intersection over
                union).
            is_aligned (bool, optional): If True, then m and n must be equal.
                Default False.

        Returns:
            Tensor: shape (m, n) if ``is_aligned `` is False else shape (m,)
        """
        assert bboxes1.size(-1) in [0, 4, 5]
        assert bboxes2.size(-1) in [0, 4, 5]
        if bboxes2.size(-1) == 5:
            bboxes2 = bboxes2[..., :4]
        if bboxes1.size(-1) == 5:
            bboxes1 = bboxes1[..., :4]

        if self.dtype == 'fp16':
            # change tensor type to save cpu and cuda memory and keep speed
            bboxes1 = cast_tensor_type(bboxes1, self.scale, self.dtype)
            bboxes2 = cast_tensor_type(bboxes2, self.scale, self.dtype)
            overlaps = bbox_overlaps(bboxes1, bboxes2, mode, is_aligned)
            if not overlaps.is_cuda and overlaps.dtype == torch.float16:
                # resume cpu float32
                overlaps = overlaps.float()
            return overlaps

        return bbox_overlaps(bboxes1, bboxes2, mode, is_aligned)

    def __repr__(self):
        """str: a string describing the module"""
        repr_str = self.__class__.__name__ + f'(' \
            f'scale={self.scale}, dtype={self.dtype})'
        return repr_str


def bbox_overlaps(bboxes1, bboxes2, mode='iou', is_aligned=False, eps=1e-6):
    """Calculate overlap between two set of bboxes.

    FP16 Contributed by https://github.com/open-mmlab/mmdetection/pull/4889
    Note:
        Assume bboxes1 is M x 4, bboxes2 is N x 4, when mode is 'iou',
        there are some new generated variable when calculating IOU
        using bbox_overlaps function:

        1) is_aligned is False
            area1: M x 1
            area2: N x 1
            lt: M x N x 2
            rb: M x N x 2
            wh: M x N x 2
            overlap: M x N x 1
            union: M x N x 1
            ious: M x N x 1

            Total memory:
                S = (9 x N x M + N + M) * 4 Byte,

            When using FP16, we can reduce:
                R = (9 x N x M + N + M) * 4 / 2 Byte
                R large than (N + M) * 4 * 2 is always true when N and M >= 1.
                Obviously, N + M <= N * M < 3 * N * M, when N >=2 and M >=2,
                           N + 1 < 3 * N, when N or M is 1.

            Given M = 40 (ground truth), N = 400000 (three anchor boxes
            in per grid, FPN, R-CNNs),
                R = 275 MB (one times)

            A special case (dense detection), M = 512 (ground truth),
                R = 3516 MB = 3.43 GB

            When the batch size is B, reduce:
                B x R

            Therefore, CUDA memory runs out frequently.

            Experiments on GeForce RTX 2080Ti (11019 MiB):

            |   dtype   |   M   |   N   |   Use    |   Real   |   Ideal   |
            |:----:|:----:|:----:|:----:|:----:|:----:|
            |   FP32   |   512 | 400000 | 8020 MiB |   --   |   --   |
            |   FP16   |   512 | 400000 |   4504 MiB | 3516 MiB | 3516 MiB |
            |   FP32   |   40 | 400000 |   1540 MiB |   --   |   --   |
            |   FP16   |   40 | 400000 |   1264 MiB |   276MiB   | 275 MiB |

        2) is_aligned is True
            area1: N x 1
            area2: N x 1
            lt: N x 2
            rb: N x 2
            wh: N x 2
            overlap: N x 1
            union: N x 1
            ious: N x 1

            Total memory:
                S = 11 x N * 4 Byte

            When using FP16, we can reduce:
                R = 11 x N * 4 / 2 Byte

        So do the 'giou' (large than 'iou').

        Time-wise, FP16 is generally faster than FP32.

        When gpu_assign_thr is not -1, it takes more time on cpu
        but not reduce memory.
        There, we can reduce half the memory and keep the speed.

    If ``is_aligned`` is ``False``, then calculate the overlaps between each
    bbox of bboxes1 and bboxes2, otherwise the overlaps between each aligned
    pair of bboxes1 and bboxes2.

    Args:
        bboxes1 (Tensor): shape (B, m, 4) in <x1, y1, x2, y2> format or empty.
        bboxes2 (Tensor): shape (B, n, 4) in <x1, y1, x2, y2> format or empty.
            B indicates the batch dim, in shape (B1, B2, ..., Bn).
            If ``is_aligned`` is ``True``, then m and n must be equal.
        mode (str): "iou" (intersection over union), "iof" (intersection over
            foreground) or "giou" (generalized intersection over union).
            Default "iou".
        is_aligned (bool, optional): If True, then m and n must be equal.
            Default False.
        eps (float, optional): A value added to the denominator for numerical
            stability. Default 1e-6.

    Returns:
        Tensor: shape (m, n) if ``is_aligned`` is False else shape (m,)

    Example:
        >>> bboxes1 = torch.FloatTensor([
        >>>     [0, 0, 10, 10],
        >>>     [10, 10, 20, 20],
        >>>     [32, 32, 38, 42],
        >>> ])
        >>> bboxes2 = torch.FloatTensor([
        >>>     [0, 0, 10, 20],
        >>>     [0, 10, 10, 19],
        >>>     [10, 10, 20, 20],
        >>> ])
        >>> overlaps = bbox_overlaps(bboxes1, bboxes2)
        >>> assert overlaps.shape == (3, 3)
        >>> overlaps = bbox_overlaps(bboxes1, bboxes2, is_aligned=True)
        >>> assert overlaps.shape == (3, )

    Example:
        >>> empty = torch.empty(0, 4)
        >>> nonempty = torch.FloatTensor([[0, 0, 10, 9]])
        >>> assert tuple(bbox_overlaps(empty, nonempty).shape) == (0, 1)
        >>> assert tuple(bbox_overlaps(nonempty, empty).shape) == (1, 0)
        >>> assert tuple(bbox_overlaps(empty, empty).shape) == (0, 0)
    """

    assert mode in ['iou', 'iof', 'giou'], f'Unsupported mode {mode}'
    # Either the boxes are empty or the length of boxes' last dimension is 4
    assert (bboxes1.size(-1) == 4 or bboxes1.size(0) == 0)
    assert (bboxes2.size(-1) == 4 or bboxes2.size(0) == 0)

    # Batch dim must be the same
    # Batch dim: (B1, B2, ... Bn)
    assert bboxes1.shape[:-2] == bboxes2.shape[:-2]
    batch_shape = bboxes1.shape[:-2]

    rows = bboxes1.size(-2)
    cols = bboxes2.size(-2)
    if is_aligned:
        assert rows == cols

    if rows * cols == 0:
        if is_aligned:
            return bboxes1.new(batch_shape + (rows, ))
        else:
            return bboxes1.new(batch_shape + (rows, cols))

    area1 = (bboxes1[..., 2] - bboxes1[..., 0]) * (
        bboxes1[..., 3] - bboxes1[..., 1])
    area2 = (bboxes2[..., 2] - bboxes2[..., 0]) * (
        bboxes2[..., 3] - bboxes2[..., 1])

    if is_aligned:
        lt = torch.max(bboxes1[..., :2], bboxes2[..., :2])  # [B, rows, 2]
        rb = torch.min(bboxes1[..., 2:], bboxes2[..., 2:])  # [B, rows, 2]

        wh = fp16_clamp(rb - lt, min=0)
        overlap = wh[..., 0] * wh[..., 1]

        if mode in ['iou', 'giou']:
            union = area1 + area2 - overlap
        else:
            union = area1
        if mode == 'giou':
            enclosed_lt = torch.min(bboxes1[..., :2], bboxes2[..., :2])
            enclosed_rb = torch.max(bboxes1[..., 2:], bboxes2[..., 2:])
    else:
        lt = torch.max(bboxes1[..., :, None, :2],
                       bboxes2[..., None, :, :2])  # [B, rows, cols, 2]
        rb = torch.min(bboxes1[..., :, None, 2:],
                       bboxes2[..., None, :, 2:])  # [B, rows, cols, 2]

        wh = fp16_clamp(rb - lt, min=0)
        overlap = wh[..., 0] * wh[..., 1]

        if mode in ['iou', 'giou']:
            union = area1[..., None] + area2[..., None, :] - overlap
        else:
            union = area1[..., None]
        if mode == 'giou':
            enclosed_lt = torch.min(bboxes1[..., :, None, :2],
                                    bboxes2[..., None, :, :2])
            enclosed_rb = torch.max(bboxes1[..., :, None, 2:],
                                    bboxes2[..., None, :, 2:])

    eps = union.new_tensor([eps])
    union = torch.max(union, eps)
    ious = overlap / union
    if mode in ['iou', 'iof']:
        return ious
    # calculate gious
    enclose_wh = fp16_clamp(enclosed_rb - enclosed_lt, min=0)
    enclose_area = enclose_wh[..., 0] * enclose_wh[..., 1]
    enclose_area = torch.max(enclose_area, eps)
    gious = ious - (enclose_area - union) / enclose_area
    return gious



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np
import torch
from torch.nn.modules.utils import _pair


def mask_target(pos_proposals_list, pos_assigned_gt_inds_list, gt_masks_list,
                cfg):
    """Compute mask target for positive proposals in multiple images.

    Args:
        pos_proposals_list (list[Tensor]): Positive proposals in multiple
            images.
        pos_assigned_gt_inds_list (list[Tensor]): Assigned GT indices for each
            positive proposals.
        gt_masks_list (list[:obj:`BaseInstanceMasks`]): Ground truth masks of
            each image.
        cfg (dict): Config dict that specifies the mask size.

    Returns:
        list[Tensor]: Mask target of each image.

    Example:
        >>> import mmcv
        >>> import mmdet
        >>> from mmdet.core.mask import BitmapMasks
        >>> from mmdet.core.mask.mask_target import *
        >>> H, W = 17, 18
        >>> cfg = mmcv.Config({'mask_size': (13, 14)})
        >>> rng = np.random.RandomState(0)
        >>> # Positive proposals (tl_x, tl_y, br_x, br_y) for each image
        >>> pos_proposals_list = [
        >>>     torch.Tensor([
        >>>         [ 7.2425,  5.5929, 13.9414, 14.9541],
        >>>         [ 7.3241,  3.6170, 16.3850, 15.3102],
        >>>     ]),
        >>>     torch.Tensor([
        >>>         [ 4.8448, 6.4010, 7.0314, 9.7681],
        >>>         [ 5.9790, 2.6989, 7.4416, 4.8580],
        >>>         [ 0.0000, 0.0000, 0.1398, 9.8232],
        >>>     ]),
        >>> ]
        >>> # Corresponding class index for each proposal for each image
        >>> pos_assigned_gt_inds_list = [
        >>>     torch.LongTensor([7, 0]),
        >>>     torch.LongTensor([5, 4, 1]),
        >>> ]
        >>> # Ground truth mask for each true object for each image
        >>> gt_masks_list = [
        >>>     BitmapMasks(rng.rand(8, H, W), height=H, width=W),
        >>>     BitmapMasks(rng.rand(6, H, W), height=H, width=W),
        >>> ]
        >>> mask_targets = mask_target(
        >>>     pos_proposals_list, pos_assigned_gt_inds_list,
        >>>     gt_masks_list, cfg)
        >>> assert mask_targets.shape == (5,) + cfg['mask_size']
    """
    cfg_list = [cfg for _ in range(len(pos_proposals_list))]
    mask_targets = map(mask_target_single, pos_proposals_list,
                       pos_assigned_gt_inds_list, gt_masks_list, cfg_list)
    mask_targets = list(mask_targets)
    if len(mask_targets) > 0:
        mask_targets = torch.cat(mask_targets)
    return mask_targets


def mask_target_single(pos_proposals, pos_assigned_gt_inds, gt_masks, cfg):
    """Compute mask target for each positive proposal in the image.

    Args:
        pos_proposals (Tensor): Positive proposals.
        pos_assigned_gt_inds (Tensor): Assigned GT inds of positive proposals.
        gt_masks (:obj:`BaseInstanceMasks`): GT masks in the format of Bitmap
            or Polygon.
        cfg (dict): Config dict that indicate the mask size.

    Returns:
        Tensor: Mask target of each positive proposals in the image.

    Example:
        >>> import mmcv
        >>> import mmdet
        >>> from mmdet.core.mask import BitmapMasks
        >>> from mmdet.core.mask.mask_target import *  # NOQA
        >>> H, W = 32, 32
        >>> cfg = mmcv.Config({'mask_size': (7, 11)})
        >>> rng = np.random.RandomState(0)
        >>> # Masks for each ground truth box (relative to the image)
        >>> gt_masks_data = rng.rand(3, H, W)
        >>> gt_masks = BitmapMasks(gt_masks_data, height=H, width=W)
        >>> # Predicted positive boxes in one image
        >>> pos_proposals = torch.FloatTensor([
        >>>     [ 16.2,   5.5, 19.9, 20.9],
        >>>     [ 17.3,  13.6, 19.3, 19.3],
        >>>     [ 14.8,  16.4, 17.0, 23.7],
        >>>     [  0.0,   0.0, 16.0, 16.0],
        >>>     [  4.0,   0.0, 20.0, 16.0],
        >>> ])
        >>> # For each predicted proposal, its assignment to a gt mask
        >>> pos_assigned_gt_inds = torch.LongTensor([0, 1, 2, 1, 1])
        >>> mask_targets = mask_target_single(
        >>>     pos_proposals, pos_assigned_gt_inds, gt_masks, cfg)
        >>> assert mask_targets.shape == (5,) + cfg['mask_size']
    """
    device = pos_proposals.device
    mask_size = _pair(cfg.mask_size)
    binarize = not cfg.get('soft_mask_target', False)
    num_pos = pos_proposals.size(0)
    if num_pos > 0:
        proposals_np = pos_proposals.cpu().numpy()
        maxh, maxw = gt_masks.height, gt_masks.width
        proposals_np[:, [0, 2]] = np.clip(proposals_np[:, [0, 2]], 0, maxw)
        proposals_np[:, [1, 3]] = np.clip(proposals_np[:, [1, 3]], 0, maxh)
        pos_assigned_gt_inds = pos_assigned_gt_inds.cpu().numpy()

        mask_targets = gt_masks.crop_and_resize(
            proposals_np,
            mask_size,
            device=device,
            inds=pos_assigned_gt_inds,
            binarize=binarize).to_ndarray()

        mask_targets = torch.from_numpy(mask_targets).float().to(device)
    else:
        mask_targets = pos_proposals.new_zeros((0, ) + mask_size)

    return mask_targets



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.cnn import CONV_LAYERS

from .builder import LINEAR_LAYERS


@LINEAR_LAYERS.register_module(name='NormedLinear')
class NormedLinear(nn.Linear):
    """Normalized Linear Layer.

    Args:
        tempeature (float, optional): Tempeature term. Default to 20.
        power (int, optional): Power term. Default to 1.0.
        eps (float, optional): The minimal value of divisor to
             keep numerical stability. Default to 1e-6.
    """

    def __init__(self, *args, tempearture=20, power=1.0, eps=1e-6, **kwargs):
        super(NormedLinear, self).__init__(*args, **kwargs)
        self.tempearture = tempearture
        self.power = power
        self.eps = eps
        self.init_weights()

    def init_weights(self):
        nn.init.normal_(self.weight, mean=0, std=0.01)
        if self.bias is not None:
            nn.init.constant_(self.bias, 0)

    def forward(self, x):
        weight_ = self.weight / (
            self.weight.norm(dim=1, keepdim=True).pow(self.power) + self.eps)
        x_ = x / (x.norm(dim=1, keepdim=True).pow(self.power) + self.eps)
        x_ = x_ * self.tempearture

        return F.linear(x_, weight_, self.bias)


@CONV_LAYERS.register_module(name='NormedConv2d')
class NormedConv2d(nn.Conv2d):
    """Normalized Conv2d Layer.

    Args:
        tempeature (float, optional): Tempeature term. Default to 20.
        power (int, optional): Power term. Default to 1.0.
        eps (float, optional): The minimal value of divisor to
             keep numerical stability. Default to 1e-6.
        norm_over_kernel (bool, optional): Normalize over kernel.
             Default to False.
    """

    def __init__(self,
                 *args,
                 tempearture=20,
                 power=1.0,
                 eps=1e-6,
                 norm_over_kernel=False,
                 **kwargs):
        super(NormedConv2d, self).__init__(*args, **kwargs)
        self.tempearture = tempearture
        self.power = power
        self.norm_over_kernel = norm_over_kernel
        self.eps = eps

    def forward(self, x):
        if not self.norm_over_kernel:
            weight_ = self.weight / (
                self.weight.norm(dim=1, keepdim=True).pow(self.power) +
                self.eps)
        else:
            weight_ = self.weight / (
                self.weight.view(self.weight.size(0), -1).norm(
                    dim=1, keepdim=True).pow(self.power)[..., None, None] +
                self.eps)
        x_ = x / (x.norm(dim=1, keepdim=True).pow(self.power) + self.eps)
        x_ = x_ * self.tempearture

        if hasattr(self, 'conv2d_forward'):
            x_ = self.conv2d_forward(x_, weight_)
        else:
            if torch.__version__ >= '1.8':
                x_ = self._conv_forward(x_, weight_, self.bias)
            else:
                x_ = self._conv_forward(x_, weight_)
        return x_



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:01
# @File    : data_collection.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from typing import Iterable, Iterator, Callable
from antgo.pipeline.functional.mixins.dag import register_dag
from antgo.pipeline.functional.mixins import DCMixins
from antgo.pipeline.functional.mixins.dataframe import DataFrameMixin
from antgo.pipeline.hparam import param_scope, dynamic_dispatch
from antgo.pipeline.functional.entity import EntityView
from antgo.pipeline.functional.option import Option, Some
from antgo.pipeline.functional.common.config import *

class DataCollection(Iterable, DCMixins):
  """A pythonic computation and processing framework.

  DataCollection is a pythonic computation and processing framework for unstructured
  data in machine learning and data science. It allows a data scientist or researcher
  to assemble data processing pipelines and do their model work (embedding,
  transforming, or classification) with a method-chaining style API. It is also
  designed to behave as a python list or iterator. When created from a list,
  operations arent performed once all data has been stored from previous step. When
  created from an iterator, operations are performed streamwise, reading and operating
  on data one by one, and only progressing if its previous output has been consumed.

  Examples:
      1. Create a DataCollection from list or iterator::

      >>> dc = DataCollection([0, 1, 2, 3, 4])
      >>> dc = DataCollection(iter([0, 1, 2, 3, 4]))

      2. Chaining function invocations makes your code clean and fluent::

      >>> (
      ...    dc.map(lambda x: x+1)
      ...      .map(lambda x: x*2)
      ... ).to_list()
      [2, 4, 6, 8, 10]

      3. Multi-line closures are also supported via decorator syntax::

      >>> dc = DataCollection([1,2,3,4])
      >>> @dc.map
      ... def add1(x):
      ...     return x+1
      >>> @add1.map
      ... def mul2(x):
      ...     return x *2
      >>> @mul2.filter
      ... def ge3(x):
      ...     return x>=7
      >>> ge3.to_list()
      [8, 10]
  """

  def __init__(self, iterable: Iterable) -> None:
    """Initializes a new DataCollection instance.

    Args:
        iterable (Iterable): The iterable data that is stored in the DataCollection.
    """
    super().__init__()
    self._iterable = iterable

  def __iter__(self) -> iter:
    """Generate an iterator of the DataCollection.

    Returns:
        iter : iterator for the data.
    """
    if hasattr(self._iterable, 'iterrows'):
      return (x[1] for x in self._iterable.iterrows())
    return iter(self._iterable)

  def __getattr__(self, name) -> 'DataCollection':
    """Unknown method dispatcher.

    When a unknown method is invoked on a `DataCollection` object, the function call
    will be dispatched to a method resolver. By registering function to the
    resolver, you are able to extend `DataCollection`'s API at runtime without
    modifying its code.

    Args:
        name (str): The unkown attribute.

    Returns:
        DataCollection: Returns a new DataCollection for the output of attribute
            call.

    Examples:
        >>> from antgo.pipeline import register
        >>> dc = DataCollection([1,2,3,4])
        >>> @register(name='test/add1')
        ... def add1(x):
        ...     return x+1
        >>> dc.test.add1().to_list()
        [2, 3, 4, 5]
    """
    if name.startswith('_'):
      return super().__getattribute__(name)

    @dynamic_dispatch
    def wrapper(*arg, **kws):
      with param_scope() as hp:
        # pylint: disable=protected-access
        path = hp._name
        index = hp._index
      
      # 添加算子节点信息到全局配置
      add_op_info(path, index, arg, kws)
      op = self.resolve(path, index, *arg, **kws)
      return self.map(op)

    return getattr(wrapper, name)

  def __getitem__(self, index) -> any:
    """Index based access of element in DataCollection.

    Access the element at the given index, similar to accessing `list[at_index]`.
    Does not work with streamed DataCollections.

    Args:
        index (int): The index location of the element being accessed.

    Raises:
        TypeError: If function called on streamed DataCollection

    Returns:
        any: The object at index.

    Examples:
        1. Usage with non-streamed::

            >>> dc = DataCollection([0, 1, 2, 3, 4])
            >>> dc[2]
            2

        2. Usage with streamed::

            >>> dc.stream()[1] # doctest: +NORMALIZE_WHITESPACE
            Traceback (most recent call last):
            TypeError: indexing is only supported for DataCollection created from list
                or pandas DataFrame.
    """
    if not hasattr(self._iterable, '__getitem__'):
      raise TypeError(
        'indexing is only supported for '
        'DataCollection created from list or pandas DataFrame.')
    if isinstance(index, int):
      return self._iterable[index]
    return DataCollection(self._iterable[index])

  def __setitem__(self, index, value):
    """Index based setting of element in DataCollection.

    Assign the value of the element at the given index, similar to
    `list[at_index]=val`. Does not work with streamed DataCollections.

    Args:
        index (int): The index location of the element being set.
        val (any): The value to be set.

    Raises:
        TypeError: If function called on streamed DataCollection

    Examples:
        1. Usage with non-streamed::

            >>> dc = DataCollection([0, 1, 2, 3, 4])
            >>> dc[2] = 3
            >>> dc.to_list()
            [0, 1, 3, 3, 4]

        2. Usage with streamed::

            >>> dc.stream()[1] # doctest: +NORMALIZE_WHITESPACE
            Traceback (most recent call last):
            TypeError: indexing is only supported for DataCollection created from list
                or pandas DataFrame.
    """
    if not hasattr(self._iterable, '__setitem__'):
      raise TypeError(
        'indexing is only supported for '
        'DataCollection created from list or pandas DataFrame.')
    self._iterable[index] = value

  @register_dag
  def __add__(self, other) -> 'DataCollection':
    """Concat two DataCollections.

    Args:
        other (DataCollection): The DataCollection being appended to the calling
            DataFrame.

    Returns:
        DataCollection: A new DataCollection of the concated DataCollections.

    Examples:
        >>> dc0 = DataCollection.range(5)
        >>> dc1 = DataCollection.range(5)
        >>> dc2 = DataCollection.range(5)
        >>> (dc0 + dc1 + dc2)
        [0, 1, 2, 3, 4, 0, ...]
    """
    self.parent_ids.append(other.id)
    other.notify_consumed(self.id)

    def inner():
      for x in self:
        yield x
      for x in other:
        yield x

    return self._factory(inner())

  def __repr__(self) -> str:
    """String representation of the DataCollection

    Returns:
        str: String repersentation of the DataCollection.

    Examples:
        1. Usage with non-streamed::

            >>> DataCollection([1, 2, 3]).unstream()
            [1, 2, 3]

        2. Usage with streamed::

            >>> DataCollection([1, 2, 3]).stream() #doctest: +ELLIPSIS
            <list_iterator object at...>
    """
    if isinstance(self._iterable, list):
      return reprlib.repr(self._iterable)
    if hasattr(self._iterable, '__repr__'):
      return repr(self._iterable)
    return super().__repr__()

  # Generation Related Function
  def _factory(self, iterable, parent_stream=True) -> 'DataCollection':
    """Factory method for Creating new DataCollections.

    This factory method has been wrapped into a `param_scope()` which contains the
    parent DataCollection's information.

    Args:
        iterable (Iterable): The data being encapsulated by the DataCollection
        parent_stream (bool, optional): Whether to use the same format of parent
            DataCollection (streamed or unstreamed). Defaults to True.

    Returns:
        DataCollection: The newly created DataCollection.
    """
    if parent_stream is True:
      if self.is_stream:
        if not isinstance(iterable, Iterator):
          iterable = iter(iterable)
      else:
        if isinstance(iterable, Iterator):
          iterable = list(iterable)

    with param_scope() as hp:
      hp().data_collection.parent = self
      return DataCollection(iterable)

  @staticmethod
  @register_dag
  def range(*arg, **kws) -> 'DataCollection':
    """Generate DataCollection with range of values.

    Generate DataCollection with a range of numbers as the data. Functions in same
    way as Python `range()` function.

    Returns:
        DataCollection: Returns a new DataCollection.

    Examples:
        >>> DataCollection.range(5).to_list()
        [0, 1, 2, 3, 4]

    """
    return DataCollection(range(*arg, **kws))

  def to_list(self) -> list:
    """Convert DataCollection to list.

    Returns:
        list: List of values stored in DataCollection.

    Examples:
        >>> DataCollection.range(5).to_list()
        [0, 1, 2, 3, 4]
    """
    return self._iterable if isinstance(self._iterable, list) else list(self._iterable)

  @register_dag
  def map(self, *arg) -> 'DataCollection':
    """Apply a function across all values in a DataCollection.

    Can apply multiple functions to the DataCollection. If multiple functions
    supplied, the same amount of new DataCollections will be returend.

    Args:
        *arg (Callable): One or multiple functions to apply to the DataCollection.

    Returns:
        DataCollection: New DataCollection containing computation results.

    Examples:
        1. Single Function::

            >>> dc = DataCollection([1,2,3,4])
            >>> dc.map(lambda x: x+1).map(lambda x: x*2).to_list()
            [4, 6, 8, 10]

        2. Multiple Functions::

            >>> dc = DataCollection([1,2,3,4])
            >>> a, b = dc.map(lambda x: x+1, lambda x: x*2)
            >>> (a.to_list(), b.to_list())
            ([2, 3, 4, 5], [2, 4, 6, 8])
    """
    # mmap
    if len(arg) > 1:
      return self.mmap(list(arg))
    unary_op = arg[0]

    if hasattr(self._iterable, 'map'):
      return self._factory(self._iterable.map(unary_op))

    if hasattr(self._iterable, 'apply') and hasattr(unary_op, '__dataframe_apply__'):
      return self._factory(unary_op.__dataframe_apply__(self._iterable))

    # map
    def inner(x):
      if isinstance(x, Option):
        return x.map(unary_op)
      else:
        return unary_op(x)

    result = map(inner, self._iterable)
    return self._factory(result)

  @register_dag
  def filter(self, unary_op: Callable, drop_empty=False) -> 'DataCollection':
    """Filter the DataCollection data based on function.

    Filters the DataCollection based on the function provided. If data is stored
    as an Option (see antgo.pipeline.functional.option.py), drop empty will decide whether
    to remove the element or set it to empty.

    Args:
        unary_op (Callable): Function that dictates filtering.
        drop_empty (bool, optional): Whether to drop empty fields. Defaults to False.

    Returns:
        DataCollection: Resulting DataCollection after filter.
    """

    def inner(x):
      if isinstance(x, Option):
        if isinstance(x, Some):
          return unary_op(x.get())
        return not drop_empty
      return unary_op(x)

    if hasattr(self._iterable, 'filter'):
      return self._factory(self._iterable.filter(unary_op))

    if hasattr(self._iterable, 'apply') and hasattr(unary_op, '__dataframe_filter__'):
      return DataCollection(unary_op.__dataframe_apply__(self._iterable))

    return self._factory(filter(inner, self._iterable))

  def run(self):
    """Iterate through the DataCollections data.

    Stream-based DataCollections will not run if the data is not a datasink. This
    function is a datasink that consumes the data without any operations.
    """
    for _ in self._iterable:
      pass

  def to_df(self) -> 'DataFrame':
    """Turn a DataCollection into a DataFrame.

    Returns:
        DataFrame: Resulting converted DataFrame.

    Examples:
        >>> from antgo.pipeline import DataCollection, Entity
        >>> e = [Entity(a=a, b=b) for a,b in zip(['abc', 'def', 'ghi'], [1,2,3])]
        >>> dc = DataCollection(e)
        >>> type(dc)
        <class 'antgo.pipeline.functional.data_collection.DataCollection'>

        >>> type(dc.to_df())
        <class 'antgo.pipeline.functional.data_collection.DataFrame'>
    """
    return DataFrame(self._iterable)


class DataFrame(DataCollection, DataFrameMixin):
  """Entity based DataCollection.

  Examples:
      >>> from antgo.pipeline import Entity
      >>> DataFrame([Entity(id=a) for a in [1,2,3]])
      [<Entity dict_keys(['id'])>, <Entity dict_keys(['id'])>, <Entity dict_keys(['id'])>]
  """

  def __init__(self, iterable: Iterable = None, **kws) -> None:
    """Initializes a new DataFrame instance.

    Args:
        iterable (Iterable, optional): The data to be encapsualted by the DataFrame.
            Defaults to None.
    """
    if iterable is not None:
      super().__init__(iterable)
      self._mode = self.ModeFlag.ROWBASEDFLAG
    else:
      super().__init__(DataFrame.from_arrow_talbe(**kws))
      self._mode = self.ModeFlag.COLBASEDFLAG

  def _factory(self, iterable, parent_stream=True, mode=None) -> 'DataFrame':
    """Factory method for Creating new DataFrames.

    This factory method has been wrapped into a `param_scope()` which contains the
    parent DataFrames's information.

    Args:
        iterable (Iterable): The data being encapsulated by the DataFrame
        parent_stream (bool, optional): Whether to use the same format of parent
            DataFrame (streamed or unstreamed). Defaults to True.
        mode (ModeFlag): The storage mode of the Dataframe.

    Returns:
        DataFrame: The newly created DataFrame.
    """

    # pylint: disable=protected-access
    if parent_stream is True:
      if self.is_stream:
        if not isinstance(iterable, Iterator):
          iterable = iter(iterable)
      else:
        if isinstance(iterable, Iterator):
          iterable = list(iterable)

    with param_scope() as hp:
      hp().data_collection.parent = self
      df = DataFrame(iterable)
      df._mode = self._mode if mode is None else mode
      return df

  def to_dc(self) -> 'DataCollection':
    """Turn a DataFrame into a DataCollection.

    Returns:
        DataCollection: Resulting DataCollection from DataFrame

    Examples:
        >>> from antgo.pipeline import DataFrame, Entity
        >>> e = [Entity(a=a, b=b) for a,b in zip(['abc', 'def', 'ghi'], [1,2,3])]
        >>> df = DataFrame(e)
        >>> type(df)
        <class 'antgo.pipeline.functional.data_collection.DataFrame'>

        >>> type(df.to_dc())
        <class 'antgo.pipeline.functional.data_collection.DataCollection'>
    """
    return DataCollection(self._iterable)

  @property
  def mode(self):
    """Storage mode of the DataFrame.

    Return the storage mode of the DataFrame.

    Returns:
        ModeFlag: The storage format of the Dataframe.

    Examples:
        >>> from antgo.pipeline import Entity, DataFrame
        >>> e = [Entity(a=a, b=b) for a,b in zip(range(5), range(5))]
        >>> df = DataFrame(e)
        >>> df.mode
        <ModeFlag.ROWBASEDFLAG: 1>

        >>> df = df.to_column()
        >>> df.mode
        <ModeFlag.COLBASEDFLAG: 2>
    """
    return self._mode

  def __iter__(self) -> iter:
    """Generate an iterator of the DataFrame.

    Returns:
        iterator: The iterator for the DataFrame.

    Examples:
        1. Row Based::

            >>> from antgo.pipeline import Entity, DataFrame
            >>> e = [Entity(a=a, b=b) for a,b in zip(range(3), range(3))]
            >>> df = DataFrame(e)
            >>> df.to_list()[0]
            <Entity dict_keys(['a', 'b'])>

        2. Column Based::

            >>> df = df.to_column()
            >>> df.to_list()[0]
            <EntityView dict_keys(['a', 'b'])>

        2. Chunk Bassed::

            >>> df = DataFrame(e)
            >>> df = df.set_chunksize(2)
            >>> df.to_list()[0]
            <EntityView dict_keys(['a', 'b'])>
    """
    if hasattr(self._iterable, 'iterrows'):
      return (x[1] for x in self._iterable.iterrows())
    if self._mode == self.ModeFlag.ROWBASEDFLAG:
      return iter(self._iterable)
    if self._mode == self.ModeFlag.COLBASEDFLAG:
      return (EntityView(i, self._iterable) for i in range(len((self._iterable))))
    if self._mode == self.ModeFlag.CHUNKBASEDFLAG:
      return (ev for wtable in self._iterable.chunks() for ev in wtable)

  def map(self, *arg) -> 'DataFrame':
    """Apply a function across all values in a DataFrame.

    Args:
        *arg (Callable): One function to apply to the DataFrame.

    Returns:
        DataFrame: New DataFrame containing computation results.
    """
    if hasattr(arg[0], '__check_init__'):
      arg[0].__check_init__()
    if self._mode == self.ModeFlag.COLBASEDFLAG or self._mode == self.ModeFlag.CHUNKBASEDFLAG:
      return self.cmap(arg[0])
    else:
      return super().map(*arg)



# -*- coding: UTF-8 -*-
# @Time    : 2019-06-08 15:51
# @File    : lsp.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
from antgo.dataflow.dataset import *
import scipy.io as io
import os
import numpy as np
import cv2
import time
from antgo.framework.helper.fileio.file_client import *


__all__ = ['MPII']

class MPII(Dataset):
    def __init__(self, train_or_test, dir=None, ext_params=None):
        super(MPII, self).__init__(train_or_test, dir, ext_params)
        self.class_name = [
            'rankl',
            'rknee',
            'rhip',
            'lhip',
            'lknee',
            'lankl',
            'pelvis',
            'thorax',
            'upper_neck',
            'head',
            'rwri',
            'relb',
            'rsho',
            'lsho',
            'lelb',
            'lwri'
        ]    
        
        if os.environ.get('LOCAL_RANK', 0) == 0:
            if not os.path.exists(os.path.join(self.dir, 'images')):
                # 数据集不存在，需要重新下载，并创建标记
                if not os.path.exists(self.dir):
                    os.makedirs(self.dir)

                ali = AliBackend()
                ali.download('ali:///dataset/mpii/mpii_human_pose_v1.tar.gz', self.dir)
                ali.download('ali:///dataset/mpii/mpii_human_pose_v1_u12_1.tar.gz', self.dir)   # 这是lspet的高精集合

                os.system(f'cd {self.dir} && unzip mpii_human_pose_v1.tar.gz')
                os.system(f'cd {self.dir} && unzip mpii_human_pose_v1_u12_1.tar.gz')
                os.system('touch DATASET_IS_READY')
            else:
                # 数据集存在，创建标记
                if not os.path.exists('DATASET_IS_READY'):
                    os.system('touch DATASET_IS_READY') 
        else:
            while True:
                # 等待直到存在指定文件
                if os.path.exists('DATASET_IS_READY'):
                    break
                time.sleep(5)

        matlab_mpii = io.loadmat(os.path.join(self.dir, 'mpii_human_pose_v1_u12_1' ,'mpii_human_pose_v1_u12_1.mat'), struct_as_record=False)['RELEASE'][0, 0]
        num_images = matlab_mpii.__dict__['annolist'][0].shape[0]

        self.info = []
        for img_idx in range(num_images):
            # Initialize empty placeholder        
            annotation_mpii = matlab_mpii.__dict__['annolist'][0, img_idx]
            train_test_mpii = matlab_mpii.__dict__['img_train'][0, img_idx].flatten()[0]
            person_id = matlab_mpii.__dict__['single_person'][img_idx][0].flatten()

            # Load the individual image. Throw an exception if image corresponding to filename not available.
            img_name = annotation_mpii.__dict__['image'][0, 0].__dict__['name'][0]
            img_path = os.path.join(self.dir, 'images', img_name)

            if self.train_or_test == 'train':
                if train_test_mpii != 1:
                    continue

            if self.train_or_test == 'test':
                if train_test_mpii == 0:
                    self.info.append({
                        'image_name': img_path,
                        'anno':{}
                    })
                else:
                    continue   

            # Iterate over persons
            joints2d = []
            joints_vis = []
            bboxes = []
            for person in (person_id - 1):
                try:
                    annopoints_img_mpii = annotation_mpii.__dict__['annorect'][0, person].__dict__['annopoints'][0, 0]
                    num_joints = annopoints_img_mpii.__dict__['point'][0].shape[0]
                    # Iterate over present joints
                    person_joints_2d = np.zeros((16, 2), dtype=np.float32)
                    person_joints_vis = np.zeros((16), dtype=np.int32)
                    person_box = np.zeros((4), dtype=np.float32)
                    for i in range(num_joints):
                        x = annopoints_img_mpii.__dict__['point'][0, i].__dict__['x'].flatten()[0]
                        y = annopoints_img_mpii.__dict__['point'][0, i].__dict__['y'].flatten()[0]
                        id_ = annopoints_img_mpii.__dict__['point'][0, i].__dict__['id'][0][0]
                        vis = annopoints_img_mpii.__dict__['point'][0, i].__dict__['is_visible'].flatten()

                        # No entry corresponding to visible
                        if vis.size == 0:
                            vis = 1
                        else:
                            vis = vis.item()

                        person_joints_2d[id_, 0] = x
                        person_joints_2d[id_, 1] = y
                        person_joints_vis[id_] = vis

                    person_box = np.array([
                        np.min(person_joints_2d[person_joints_vis, 0]), 
                        np.min(person_joints_2d[person_joints_vis, 1]), 
                        np.max(person_joints_2d[person_joints_vis, 0]), 
                        np.max(person_joints_2d[person_joints_vis, 1])
                        ])
                    bboxes.append(person_box)
                    joints2d.append(person_joints_2d)
                    joints_vis.append(person_joints_vis)
                    
                except KeyError:
                    # Person 'x' could not have annotated joints, hence move to person 'y'
                    continue

            if len(joints2d) == 0:
                continue

            anno = {
                'bboxes': np.stack(bboxes, 0),
                'labels': np.zeros((len(bboxes)), dtype=np.int32),
                'joints2d': np.stack(joints2d, 0),
                'joints_vis': np.stack(joints_vis, 0)
            }
            self.info.append({
                'image_name': img_path,
                'anno': anno
            })

    @property
    def size(self):
        return len(self.info)

    def at(self, img_idx):
        image = cv2.imread(self.info[img_idx]['image_name'])
        return (image, self.info[img_idx]['anno'])

    def split(self, split_params={}, split_method=''):
        raise NotImplementedError

# mpii = MPII('train', '/root/workspace/dataset/mpii')
# for i in range(1, mpii.size):
#     data = mpii.sample(i)
#     image = data['image']
#     joints2d = data['joints2d']
#     joints_vis = data['joints_vis']
    
#     for joint_i, (x,y) in enumerate(joints2d[0]):
#         x, y = int(x), int(y)
        
#         if joints_vis[0][joint_i]:
#             cv2.circle(image, (x, y), radius=2, color=(0,0,255), thickness=1)

#     cv2.imwrite(f'./aabb_{i}.png', image)
#     print(i)


from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.framework.helper.models.proxy_module import ProxyModule
from antgo.framework.helper.runner.test import single_gpu_test

import os
import os.path as osp
import math
import time
import glob
import abc
from torch.utils.data import DataLoader, dataloader
import torch.optim
import torchvision.transforms as transforms
from torch.nn.parallel.data_parallel import DataParallel
from antgo.framework.helper.utils.config import Config
from antgo.framework.helper.apis.train import *
from antgo.framework.helper.dataset import (build_dataset, build_dataloader, build_kv_dataloader, build_iter_dataloader)
from antgo.framework.helper.utils.util_distribution import build_ddp, build_dp, get_device
from antgo.framework.helper.utils import get_logger
from antgo.framework.helper.runner import *
from antgo.framework.helper.runner import get_dist_info, init_dist
from antgo.framework.helper.runner.builder import *
from antgo.framework.helper.utils.setup_env import *
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.utils import *
from antgo.framework.helper.models.dummy_module import *
from antgo.framework.helper.models.proxy_module import *
from antgo.framework.helper.models.distillation import *
from antgo.framework.helper.runner.hooks.hook import *
from antgo.framework.helper.parallel.utils import is_module_wrapper
from .base_trainer import *
from thop import profile
import copy
import logging
import tempfile

import torch.distributed as dist
from contextlib import contextmanager
import json

try:
    from aimet_common.defs import QuantScheme
    from aimet_torch import bias_correction
    from aimet_torch.quantsim import QuantParams, QuantizationSimModel
    from aimet_torch.cross_layer_equalization import equalize_model
    from aimet_torch.adaround.adaround_weight import Adaround, AdaroundParameters
    from aimet_torch.batch_norm_fold import fold_all_batch_norms
    from aimet_torch.onnx_utils import OnnxExportApiArgs
    from aimet_torch.model_preparer import prepare_model
    from aimet_common.utils import start_bokeh_server_session
except:
    print('Dont support aimet.')
    pass


def _default_forward_pass_callback(model, data_loader):
    # switch to eval state
    model.eval()

    iteration = 10
    prog_bar = ProgressBar(iteration)
    count = 0
    for data in data_loader:
        data = data.to('cuda')
        with torch.no_grad():
            model(data)

        prog_bar.update()
        count += 1
        if count >= iteration:
            break

def apply_cross_layer_equalization(model, input_shape):
    """
    Applies CLE on the model and calculates model accuracy on quantized simulator
    Applying CLE on the model inplace consists of:
        Batch Norm Folding
        Cross Layer Scaling
        High Bias Fold
    Converts any ReLU6 into ReLU.
    :param model: the loaded model
    :param input_shape: the shape of the input to the model
    :return:
    """

    equalize_model(model, input_shape)


def apply_bias_correction(model, data_loader):
    """
    Applies Bias-Correction on the model.
    :param model: The model to quantize
    :param dataloader: DataLoader used during quantization
    :param logdir: Log directory used for storing log files
    :return: None
    """
    # Rounding mode can be 'nearest' or 'stochastic'
    rounding_mode = 'nearest'

    # Number of samples used during quantization
    num_quant_samples = 256*20

    # Number of samples used for bias correction
    num_bias_correct_samples = 256*20

    params = QuantParams(weight_bw=8, act_bw=8, round_mode=rounding_mode, quant_scheme='tf_enhanced')

    # Perform Bias Correction
    bias_correction.correct_bias(model.to(device="cuda"), params, num_quant_samples=num_quant_samples,
                                 data_loader=data_loader, num_bias_correct_samples=num_bias_correct_samples)


def calculate_quantsim(model, val_dataloader, dummy_input, use_cuda, path, prefix):
    """
    Calculates model accuracy on quantized simulator and returns quantized model with accuracy.
    :param model: the loaded model
    :param evaluator: the Eval function to use for evaluation
    :param iterations: No of batches to use in computing encodings.
                       Not used in image net dataset
    :param num_val_samples_per_class: No of samples to use from every class in
                                      computing encodings. Not used in pascal voc
                                      dataset
    :param use_cuda: the cuda device.
    :return: a tuple of quantsim and accuracy of model on this quantsim
    """
    # Number of batches to use for computing encodings
    # Only 5 batches are used here to speed up the process, also the
    # number of images in these 5 batches should be sufficient for
    # compute encodings
    quantsim = QuantizationSimModel(model=model, quant_scheme='tf_enhanced',
                                    dummy_input=dummy_input, rounding_mode='nearest',
                                    default_output_bw=8, default_param_bw=8, in_place=False)

    quantsim.compute_encodings(forward_pass_callback=_default_forward_pass_callback,
                               forward_pass_callback_args=val_dataloader)

    if not os.path.exists(path):
        os.makedirs(path)
    quantsim.export(path=path, filename_prefix=prefix, dummy_input=dummy_input.cpu())
    return quantsim

class Trainer(BaseTrainer):
    def __init__(self, cfg, work_dir="./", gpu_id=-1, distributed=False, diff_seed=True, deterministic=True, find_unused_parameters=False):
        if isinstance(cfg, dict):
            self.cfg = Config.fromstring(json.dumps(cfg), '.json')
        else:
            self.cfg = cfg
            
        self.data_loaders = None
        self.runner = None
        self.work_dir = work_dir
        self.train_generator = None
        self.val_dataloader = None
        self.distributed = distributed
        self.find_unused_parameters = find_unused_parameters
        self.meta = {}

        # set multi-process settings
        setup_multi_processes(self.cfg)

        gpu_id = int(gpu_id)
        device = 'cpu' if gpu_id < 0 else 'cuda'
        self.cfg.gpu_ids = [gpu_id] if gpu_id >= 0 else []
        if self.distributed:
            init_dist(**self.cfg.get('dist_params', {}))
            # re-set gpu_ids with distributed training mode
            _, world_size = get_dist_info()
            self.cfg.gpu_ids = range(world_size)
        
        # set random seeds
        seed = init_random_seed(self.cfg.get('seed', 0), device=device)
        seed = seed + dist.get_rank() if diff_seed else seed
        set_random_seed(seed, deterministic=deterministic)
        self.cfg.seed = seed
        self.meta['seed'] = seed
        self.device = device

    def config_dataloader(self, with_validate=False):
        # 创建数据集
        dataset = build_dataset(self.cfg.data.train)

        # 创建数据集加载器
        train_dataloader_default_args = dict(
            samples_per_gpu=2,
            workers_per_gpu=2,
            # `num_gpus` will be ignored if distributed
            num_gpus=1,
            dist=self.distributed,
            seed=self.cfg.seed,
            runner_type='EpochBasedRunner',
            persistent_workers=True)

        train_loader_cfg = {
            **train_dataloader_default_args,
            **self.cfg.data.get('train_dataloader', {})
        }

        if getattr(dataset, 'is_kv', False):
            self.train_generator = build_kv_dataloader(dataset, **train_loader_cfg)
        elif isinstance(dataset, torch.utils.data.IterableDataset):
            self.train_generator = build_iter_dataloader(dataset, **train_loader_cfg)
        else:
            self.train_generator = build_dataloader(dataset, **train_loader_cfg)

        if with_validate:
            val_dataloader_default_args = dict(
                samples_per_gpu=1,
                workers_per_gpu=2,
                dist=self.distributed,
                shuffle=False,
                persistent_workers=True)
            
            val_dataloader_args = {
                    **val_dataloader_default_args,
                    **self.cfg.data.get('val_dataloader', {})
            }
            val_dataset = build_dataset(self.cfg.data.val)

            if getattr(dataset, 'is_kv', False):
                self.val_dataloader = build_kv_dataloader(val_dataset, **val_dataloader_args)
            elif isinstance(dataset, torch.utils.data.IterableDataset):
                val_dataset.is_infinite = False
                self.val_dataloader = build_iter_dataloader(val_dataset, **val_dataloader_args)
            else:
                self.val_dataloader = build_dataloader(val_dataset, **val_dataloader_args)

    def config_model(self, model_builder=None, resume_from=None, load_from=None, revise_keys=[(r'^module\.', '')]):
        # prepare network
        logger = get_logger('model', log_level=self.cfg.get('log_level', logging.INFO))
        logger.info("Creating graph and optimizer...")

        # 构建网络
        if model_builder is not None:
            model = model_builder()
        else:
            model = build_model(self.cfg.model)

        # 模型初始化
        model.init_weights()

        if self.distributed:
            if self.cfg.get('syncBN', True):
                model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model).to(self.device)

            model = build_ddp(
                model,
                self.device,
                device_ids=[int(os.environ['LOCAL_RANK'])],
                broadcast_buffers=False,
                find_unused_parameters=self.find_unused_parameters)
        else:
            model = build_dp(model, self.device, device_ids=self.cfg.gpu_ids)

        # 自动调整单设备下的学习率到多设备下的学习率
        auto_scale_lr(self.cfg, self.distributed, logger)

        # build optimizer
        # 如果构造optimizer调度，则不在全局hook中注册
        optimizer = None
        if 'type' in self.cfg.optimizer:
            # optimizer 简单，全局控制
            optimizer = build_optimizer(model, self.cfg.optimizer)
        else:
            # optimizer 复合，对多个子模块独立控制
            # 需要模型自己控制
            optimizer = {}
            if is_module_wrapper(model):
                model = model.module

            for submodule_name, optimizer_dict in self.cfg.optimizer.items():
                optimizer[submodule_name] = build_optimizer(getattr(model, submodule_name), optimizer_dict)

        # build lr scheduler
        # 如果构造复合lr调度，则不在全局hook中注册
        lr_scheduler = {}
        if 'policy' not in self.cfg.lr_config:
            # lr schedule 复合，对多个子模块独立控制
            for submodule_name, lr_config in self.cfg.lr_config.items():
                assert('policy' in lr_config)

                policy_type = lr_config.pop('policy')
                # If the type of policy is all in lower case, e.g., 'cyclic',
                # then its first letter will be capitalized, e.g., to be 'Cyclic'.
                # This is for the convenient usage of Lr updater.
                # Since this is not applicable for `
                # CosineAnnealingLrUpdater`,
                # the string will not be changed if it contains capital letters.
                if policy_type == policy_type.lower():
                    policy_type = policy_type.title()
                hook_type = policy_type + 'LrUpdaterHook'
                lr_config['type'] = hook_type
                lr_scheduler[submodule_name] = build_from_cfg(lr_config, HOOKS)         

        # build training strategy
        custom_runner = self.cfg.get('runner', dict(type='EpochBasedRunner', max_epochs=1))
        self.runner = build_runner(
            custom_runner,        # 忽略max_epochs，开发者控制最大epoch
            default_args=dict(
                model=model,
                optimizer=optimizer,
                work_dir=self.work_dir,
                logger=logger,
                meta=self.meta,
                lr_scheduler=lr_scheduler))

        # an ugly workaround to make .log and .log.json filenames the same
        self.runner.timestamp = time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())

        optimizer_config = None
        if not isinstance(optimizer, dict):
            # 对于复合optimizer，不进行后台hook
            # 对于optimizer的控制由模型的train_step内部自己控制
            optimizer_config = self.cfg.get('optimizer_config', {})
            if 'type' not in optimizer_config:
                optimizer_config = OptimizerHook(**optimizer_config)

        custom_hooks = self.cfg.get('custom_hooks', None)        
        self.runner.register_training_hooks(
            self.cfg.lr_config if len(lr_scheduler) == 0 else None,     # 学习率调整策略，比如step,warmup等
            optimizer_config,                                           # 优化器的相关后处理，比如限制梯度操作等
            self.cfg.checkpoint_config,                                 # checkpoint相关处理
            self.cfg.log_config,                                         
            self.cfg.get('momentum_config', None),                       
            custom_hooks_config=custom_hooks)

        if self.distributed:
            if isinstance(self.runner, EpochBasedRunner):
                self.runner.register_hook(DistSamplerSeedHook())

        if self.val_dataloader is not None:
            eval_cfg = self.cfg.get('evaluation', {})
            eval_cfg['by_epoch'] = True
            if 'interval' not in eval_cfg:
                eval_cfg['interval'] = 1

            if 'metric' in eval_cfg:
                metric = build_measure(eval_cfg['metric'])
                eval_cfg['metric'] = metric

            if 'type' not in eval_cfg:
                eval_hook = DistEvalHook if self.distributed else EvalHook
                self.runner.register_hook(
                    eval_hook(self.val_dataloader, **eval_cfg), priority='LOW')
            else:
                eval_cfg['dataloader'] = self.val_dataloader
                self.runner.register_hook(
                    build_from_cfg(eval_cfg, HOOKS), priority='LOW'
                )

        if resume_from:
            self.runner.resume(resume_from)
        elif load_from:
            self.runner.load_checkpoint(load_from, revise_keys=revise_keys)
        else:
            if self.distributed:
                # 为了稳妥起见，把rank0的权重保存下来，其余进程进行加载保证所有卡的起始权重相同
                checkpoint_path = os.path.join(tempfile.gettempdir(), "initial_weights.pt")
                rank, _ = get_dist_info()
                if rank == 0:
                    torch.save(model.state_dict(), checkpoint_path)

                dist.barrier()
                self.runner.load_checkpoint(checkpoint_path)
        
    def apply_ptq_quant(self, dummy_input, checkpoint, model_builder=None, path='./', prefix='quant'):
        ###############################     STEP - 0    ###############################
        # 计算浮点模型
        logger = get_logger('qat', log_level=self.cfg.log_level)
        if model_builder is not None:
            f32_model = model_builder()
        else:
            f32_model = build_model(self.cfg.model)

        assert(dummy_input is not None)
        # 加载checkpoint
        ckpt = torch.load(checkpoint, map_location='cpu')
        f32_model.load_state_dict(ckpt['state_dict'], strict=True)

        # 计算浮点模型精度
        print('Computing Float Model Accuray.')
        eval_cfg = self.cfg.get('evaluation', {})
        metric_func = None
        if 'metric' in eval_cfg:
            metric_func = build_measure(eval_cfg['metric'])

        warp_model = build_dp(f32_model, self.device, device_ids=self.cfg.gpu_ids)  
        results = single_gpu_test(warp_model, self.val_dataloader)    
        eval_res = {}
        if metric_func is None:
            eval_res = self.val_dataloader.dataset.evaluate(results, logger=None)
        else:
            gts = []
            for gt_i in range(len(self.val_dataloader.dataset)):
                gts.append(self.val_dataloader.dataset.get_ann_info(gt_i))

            eval_res = metric_func(results, gts)                 
        print(eval_res)

        # 获得浮点模型的 FLOPS、PARAMS
        model = DummyModelWarp(f32_model)        
        model = model.to('cpu')
        dummy_input = dummy_input.to('cpu')
        flops, params = profile(model, inputs=(dummy_input,))
        print('FLOPs = ' + str(flops/1000**3) + 'G')
        print('Params = ' + str(params/1000**2) + 'M')

        if not os.path.exists(os.path.join(path, 'float')):
            os.makedirs(os.path.join(path, 'float'))
        # Export the model
        torch.onnx.export(
                model,                                      # model being run
                dummy_input,                                # model input (or a tuple for multiple inputs)
                os.path.join(path, 'float', 'model.onnx'),  # where to save the model (can be a file or file-like object)
                export_params=True,                         # store the trained parameter weights inside the model file
                opset_version=11,                           # the ONNX version to export the model to
                do_constant_folding=True,                   # whether to execute constant folding for optimization
                input_names = ['input'],   # the model's input names
        )

        # Quantize the model using AIMET QAT (quantization aware training) and calculate accuracy on Quant Simulator
        ###############################     STEP - 1    ###############################
        print('Computing Naive Quantizing Progress.')
        model = model.to(self.device)
        quant_val_loader =  \
            torch.utils.data.DataLoader(UnlabeledDatasetWrapper(self.val_dataloader.dataset, 'image'),
                                        batch_size=128, 
                                        shuffle=True)     
        dummy_input = dummy_input.to(self.device)
        quant_sim = calculate_quantsim(model, quant_val_loader, dummy_input, True, os.path.join(path, 'NAIVE'), prefix)

        print("Naive Quantized Model performance")
        # self.cfg.qat.update(dict(feature_func=quant_sim.model))
        # warp_model = build_model(self.cfg.qat) 
        head_cfg = None
        for k in self.cfg.model.keys():
            if 'head' in k:
                head_cfg = self.cfg.model[k]
        assert(head_cfg is not None)
        proxy_model_cfg = self.cfg['proxy_model'] if 'proxy_model' in self.cfg.keys() else {}
        warp_model = ProxyModule(quant_sim.model, head_cfg, init_cfg=proxy_model_cfg['init_cfg'] if 'init_cfg' in proxy_model_cfg else None)
        warp_model = build_dp(warp_model, self.device, device_ids=self.cfg.gpu_ids)        
        results = single_gpu_test(warp_model, self.val_dataloader)            
        eval_res = {}
        if metric_func is None:
            eval_res = self.val_dataloader.dataset.evaluate(results, logger=None)
        else:
            gts = []
            for gt_i in range(len(self.val_dataloader.dataset)):
                gts.append(self.val_dataloader.dataset.get_ann_info(gt_i))

            eval_res = metric_func(results, gts)                 
        print(eval_res)        
        
        ###############################     STEP - 2    ###############################
        # For good initialization apply, apply Post Training Quantization (PTQ) methods
        # such as Cross Layer Equalization (CLE) and Bias Correction (BC) (optional)
        print('Computing PTQ(CLE and BC) Quantizing Progress.')
        apply_cross_layer_equalization(model=model, input_shape=dummy_input.shape)
        apply_bias_correction(model=model, 
                              data_loader=torch.utils.data.DataLoader(LabeledDatasetWrapper(self.val_dataloader.dataset, 'image'), batch_size=256, shuffle=True))
        dummy_input = dummy_input.to(self.device)
        quant_sim = calculate_quantsim(model, quant_val_loader, dummy_input, True, os.path.join(path, 'PTQ'), prefix)
        
        print('Post Training Quantization (PTQ) Complete')
        # self.cfg.qat.update(dict(feature_func=quant_sim.model))
        # warp_model = build_model(self.cfg.qat) 
        warp_model = ProxyModule(quant_sim.model, head_cfg, init_cfg=proxy_model_cfg['init_cfg'] if 'init_cfg' in proxy_model_cfg else None)
        warp_model = build_dp(warp_model, self.device, device_ids=self.cfg.gpu_ids)
        results = single_gpu_test(warp_model, self.val_dataloader)  
        eval_res = {}
        if metric_func is None:
            eval_res = self.val_dataloader.dataset.evaluate(results, logger=None)
        else:
            gts = []
            for gt_i in range(len(self.val_dataloader.dataset)):
                gts.append(self.val_dataloader.dataset.get_ann_info(gt_i))

            eval_res = metric_func(results, gts)                 
        print(eval_res)           

    def apply_qat_quant(self, dummy_input, checkpoint, model_builder=None, path='./', prefix='quant'):
        logger = get_logger('qat', log_level=self.cfg.log_level)
        if model_builder is not None:
            f32_model = model_builder()
        else:
            f32_model = build_model(self.cfg.model)

        assert(dummy_input is not None)
        # 加载checkpoint
        ckpt = torch.load(checkpoint, map_location='cpu')
        f32_model.load_state_dict(ckpt['state_dict'], strict=True)

        model = DummyModelWarp(f32_model)        
        model = model.to(self.device)
        quant_val_loader =  \
            torch.utils.data.DataLoader(
                UnlabeledDatasetWrapper(self.val_dataloader.dataset, 'image'),batch_size=128, 
                                        shuffle=False)    

        ###############################     STEP - 0    ###############################
        # For good initialization apply, apply Post Training Quantization (PTQ) methods
        # such as Cross Layer Equalization (CLE) and Bias Correction (BC) (optional)
        if not self.distributed or (get_dist_info()[0] == 0):
            print('Computing PTQ(CLE and BC) Quantizing Progress.')
        apply_cross_layer_equalization(model=model, input_shape=dummy_input.shape)
        apply_bias_correction(model=model, 
                              data_loader=torch.utils.data.DataLoader(
                                  LabeledDatasetWrapper(self.val_dataloader.dataset, 'image'), 
                                  batch_size=256, 
                                  shuffle=True))
        dummy_input = dummy_input.to(self.device)
        quant_sim = calculate_quantsim(model, quant_val_loader, dummy_input, True, os.path.join(path, 'PTQ'), prefix)

        # 删除原始模型      
        del model

        if not self.distributed or (get_dist_info()[0] == 0):
            print('Post Training Quantization (PTQ) Complete')
        head_cfg = None
        for k in self.cfg.model.keys():
            if 'head' in k:
                head_cfg = self.cfg.model[k]
        assert(head_cfg is not None)            
        proxy_model_cfg = self.cfg['proxy_model'] if 'proxy_model' in self.cfg.keys() else {}        
        model = ProxyModule(quant_sim.model, head_cfg, init_cfg=proxy_model_cfg['init_cfg'] if 'init_cfg' in proxy_model_cfg else None)
        
        if self.distributed:
            find_unused_parameters = self.cfg.get('find_unused_parameters', False)
            # Sets the `find_unused_parameters` parameter in
            # torch.nn.parallel.DistributedDataParallel
            model = build_ddp(
                model,
                self.device,
                device_ids=[int(os.environ['LOCAL_RANK'])],
                broadcast_buffers=False,
                find_unused_parameters=find_unused_parameters)
        else:
            model = build_dp(model, self.device, device_ids=self.cfg.gpu_ids)

        ###############################     STEP - 1    ###############################
        # Finetune the quantized model
        if not self.distributed or (get_dist_info()[0] == 0):        
            print('Starting Model Finetune')
        # build optimizer        
        optimizer = build_optimizer(model, self.cfg.optimizer)

        # 自动调整单设备下的学习率到多设备下的学习率
        auto_scale_lr(self.cfg, self.distributed, logger)

        # build training strategy
        self.runner = build_runner(
            dict(type='EpochBasedRunner', max_epochs=1),        # 忽略max_epochs，开发者控制最大epoch
            default_args=dict(
                model=model,
                optimizer=optimizer,
                work_dir=self.work_dir,
                logger=logger,
                meta=self.meta))

        # an ugly workaround to make .log and .log.json filenames the same
        self.runner.timestamp = time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())

        optimizer_config = self.cfg.optimizer_config
        if self.distributed and 'type' not in self.cfg.optimizer_config:
            optimizer_config = OptimizerHook(**self.cfg.optimizer_config)

        self.runner.register_training_hooks(
            self.cfg.lr_config,                                     # 学习率调整策略，比如step,warmup等
            optimizer_config,                                       # 优化器的相关后处理，比如限制梯度操作等
            self.cfg.checkpoint_config,                             # checkpoint相关处理
            self.cfg.log_config,                                         
            self.cfg.get('momentum_config', None),                       
            custom_hooks_config=self.cfg.get('custom_hooks', None))

        if self.distributed:
            if isinstance(self.runner, EpochBasedRunner):
                self.runner.register_hook(DistSamplerSeedHook())

        # 配置验证方法
        eval_cfg = self.cfg.get('evaluation', {})
        eval_cfg['by_epoch'] = True
        if 'interval' not in eval_cfg:
            eval_cfg['interval'] = 1

        eval_hook = DistEvalHook if self.distributed else EvalHook

        if 'metric' in eval_cfg:
            metric = build_measure(eval_cfg['metric'])
            eval_cfg['metric'] = metric
        self.runner.register_hook(
            eval_hook(self.val_dataloader, **eval_cfg), priority='LOW')

        # 训练
        with self.train_context(15) as runner:
            for _ in range(0, runner._max_epochs):
                self.run_on_train()

        # Save the quantized model
        # only on rank=0 node
        if not self.distributed or (get_dist_info()[0] == 0):
            if not os.path.exists(os.path.join(path, 'QAT')):
                os.makedirs(os.path.join(path, 'QAT'))
            quant_sim.export(path=os.path.join(path, 'QAT'), filename_prefix=prefix, dummy_input=dummy_input.cpu())


# -*- coding: UTF-8 -*-
# @Time    : 2022/4/27 20:38
# @File    : my_seg_dataset.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import gzip
import random
import numpy as np
from .dataset import Dataset
from antgo.dataflow.imgaug.operators_3d import *
from antgo.dataflow.vis import *
from pycocotools.coco import COCO
import time
import copy
import json


class InterHand26M(Dataset):
    def __init__(self, train_or_test, dir=None, params=None):
        super(InterHand26M, self).__init__(train_or_test, dir, params)
    
        self.mode = train_or_test # train, test, val
        self.img_path = os.path.join(dir, 'images')
        self.annot_path = os.path.join(dir, 'annotations')
        if self.mode == 'val':
            self.rootnet_output_path = os.path.join(dir,'rootnet_output/rootnet_interhand2.6m_output_val.json')
        else:
            self.rootnet_output_path = os.path.join(dir,'rootnet_output/rootnet_interhand2.6m_output_test.json' )

        self.joint_num = 21 # single hand
        self.root_joint_idx = {'right': 20, 'left': 41}
        self.joint_type = {'right': np.arange(0,self.joint_num), 'left': np.arange(self.joint_num,self.joint_num*2)}
        self.skeleton = load_skeleton(osp.join(self.annot_path, 'skeleton.txt'), self.joint_num*2)
        
        self.datalist = []
        self.datalist_sh = []
        self.datalist_ih = []
        self.sequence_names = []
        
        # load annotation
        print("Load annotation from  " + osp.join(self.annot_path, self.mode))
        db = COCO(osp.join(self.annot_path, self.mode, 'InterHand2.6M_' + self.mode + '_data.json'))
        with open(osp.join(self.annot_path, self.mode, 'InterHand2.6M_' + self.mode + '_camera.json')) as f:
            cameras = json.load(f)
        with open(osp.join(self.annot_path, self.mode, 'InterHand2.6M_' + self.mode + '_joint_3d.json')) as f:
            joints = json.load(f)
        
        trans_test = getattr(self, 'trans_test', '')
        if (self.mode == 'val' or self.mode == 'test') and trans_test == 'rootnet':
            print("Get bbox and root depth from " + self.rootnet_output_path)
            rootnet_result = {}
            with open(self.rootnet_output_path) as f:
                annot = json.load(f)
            for i in range(len(annot)):
                rootnet_result[str(annot[i]['annot_id'])] = annot[i]
        else:
            print("Get bbox and root depth from groundtruth annotation")
        
        for aid in db.anns.keys():
            ann = db.anns[aid]
            image_id = ann['image_id']
            img = db.loadImgs(image_id)[0]
 
            capture_id = img['capture']
            seq_name = img['seq_name']
            cam = img['camera']
            frame_idx = img['frame_idx']
            img_path = osp.join(self.img_path, self.mode, img['file_name'])
            
            campos, camrot = np.array(cameras[str(capture_id)]['campos'][str(cam)], dtype=np.float32), np.array(cameras[str(capture_id)]['camrot'][str(cam)], dtype=np.float32)
            focal, princpt = np.array(cameras[str(capture_id)]['focal'][str(cam)], dtype=np.float32), np.array(cameras[str(capture_id)]['princpt'][str(cam)], dtype=np.float32)
            joint_world = np.array(joints[str(capture_id)][str(frame_idx)]['world_coord'], dtype=np.float32)
            joint_cam = world2cam(joint_world.transpose(1,0), camrot, campos.reshape(3,1)).transpose(1,0)
            joint_img = cam2pixel(joint_cam, focal, princpt)[:,:2]

            joint_valid = np.array(ann['joint_valid'],dtype=np.float32).reshape(self.joint_num*2)
            # if root is not valid -> root-relative 3D pose is also not valid. Therefore, mark all joints as invalid
            joint_valid[self.joint_type['right']] *= joint_valid[self.root_joint_idx['right']]
            joint_valid[self.joint_type['left']] *= joint_valid[self.root_joint_idx['left']]
            hand_type = ann['hand_type']
            hand_type_valid = np.array((ann['hand_type_valid']), dtype=np.float32)
            
            if (self.mode == 'val' or self.mode == 'test') and trans_test == 'rootnet':
                bbox = np.array(rootnet_result[str(aid)]['bbox'],dtype=np.float32)
                abs_depth = {'right': rootnet_result[str(aid)]['abs_depth'][0], 'left': rootnet_result[str(aid)]['abs_depth'][1]}
            else:
                img_width, img_height = img['width'], img['height']
                bbox = np.array(ann['bbox'],dtype=np.float32) # x,y,w,h
                input_img_shape = getattr(self, 'input_img_shape', None)
                bbox = process_bbox(bbox, (img_height, img_width), input_img_shape)
                abs_depth = {'right': joint_cam[self.root_joint_idx['right'],2], 'left': joint_cam[self.root_joint_idx['left'],2]}

            cam_param = {'focal': focal, 'princpt': princpt}
            joint = {'cam_coord': joint_cam, 'img_coord': joint_img, 'valid': joint_valid}
            data = {'img_path': img_path, 'seq_name': seq_name, 'cam_param': cam_param, 'bbox': bbox, 'joint': joint, 'hand_type': hand_type, 'hand_type_valid': hand_type_valid, 'abs_depth': abs_depth, 'file_name': img['file_name'], 'capture': capture_id, 'cam': cam, 'frame': frame_idx}
            if hand_type == 'right' or hand_type == 'left':
                self.datalist_sh.append(data)
            else:
                self.datalist_ih.append(data)
            
            if seq_name not in self.sequence_names:
                self.sequence_names.append(seq_name)

        self.datalist = self.datalist_sh + self.datalist_ih
        self._set_group_flag()
        print('Number of annotations in single hand sequences: ' + str(len(self.datalist_sh)))
        print('Number of annotations in interacting hand sequences: ' + str(len(self.datalist_ih)))

    def handtype_str2array(self, hand_type):
        if hand_type == 'right':
            return np.array([1,0], dtype=np.float32)
        elif hand_type == 'left':
            return np.array([0,1], dtype=np.float32)
        elif hand_type == 'interacting':
            return np.array([1,1], dtype=np.float32)
        else:
            assert 0, print('Not supported hand type: ' + hand_type)
    
    @property
    def size(self):
        return len(self.datalist)

    def at(self, idx):
        data = self.datalist[idx]
        img_path, bbox, joint, hand_type, hand_type_valid = data['img_path'], data['bbox'], data['joint'], data['hand_type'], data['hand_type_valid']
        joint_cam = joint['cam_coord'].copy(); joint_img = joint['img_coord'].copy(); joint_valid = joint['valid'].copy();
        hand_type = self.handtype_str2array(hand_type)
        joint_coord = np.concatenate((joint_img, joint_cam[:,2,None]),1)
        
        # image load
        img = load_img(img_path)
        # augmentation
        input_img_shape = getattr(self, 'input_img_shape', None)
        output_hm_shape = getattr(self, 'output_hm_shape', '')
        bbox_3d_size = getattr(self, 'bbox_3d_size', 0)
        output_root_hm_shape = getattr(self, 'output_root_hm_shape', 0)
        bbox_3d_size_root = getattr(self, 'bbox_3d_size_root', 0)
        img, joint_coord, joint_valid, hand_type, inv_trans = augmentation(img, bbox, joint_coord, joint_valid, hand_type, self.mode, self.joint_type, input_img_shape)
        rel_root_depth = np.array([joint_coord[self.root_joint_idx['left'],2] - joint_coord[self.root_joint_idx['right'],2]],dtype=np.float32).reshape(1)
        root_valid = np.array([joint_valid[self.root_joint_idx['right']] * joint_valid[self.root_joint_idx['left']]],dtype=np.float32).reshape(1) if hand_type[0]*hand_type[1] == 1 else np.zeros((1),dtype=np.float32)
        # transform to output heatmap space
        joint_coord, joint_valid, rel_root_depth, root_valid = transform_input_to_output_space(joint_coord, joint_valid, rel_root_depth, root_valid, self.root_joint_idx, self.joint_type, input_img_shape, output_hm_shape, bbox_3d_size, output_root_hm_shape, bbox_3d_size_root)
        img = img.astype(np.float32) / 255.
        
        inputs = {'image': img}
        targets = {'joint_coord': joint_coord, 'rel_root_depth': rel_root_depth, 'hand_type': hand_type}
        meta_info = {'joint_valid': joint_valid, 'root_valid': root_valid, 'hand_type_valid': hand_type_valid, 'inv_trans': inv_trans, 'capture': int(data['capture']), 'cam': int(data['cam']), 'frame': int(data['frame'])}
        annotation = {}
        annotation.update(inputs)
        annotation.update(targets)
        annotation.update(meta_info)
        return None, annotation

    def data_pool(self):
        for index in range(self.size):
            yield self.at(index)

    def _set_group_flag(self):
        """Set flag according to image aspect ratio.

        Images with aspect ratio greater than 1 will be set as group 1,
        otherwise group 0.
        """
        self.flag = np.zeros(self.size, dtype=np.uint8)

    def evaluate(self, preds, **kwargs):
        print('Evaluation start...')

        gts = self.datalist
        # preds_joint_coord, preds_rel_root_depth, preds_hand_type, inv_trans = preds['joint_coord'], preds['rel_root_depth'], preds['hand_type'], preds['inv_trans']

        # assert len(gts) == len(preds_joint_coord)
        sample_num = len(gts)
        
        mpjpe_sh = [[] for _ in range(self.joint_num*2)]
        mpjpe_ih = [[] for _ in range(self.joint_num*2)]
        mrrpe = []
        acc_hand_cls = 0; hand_cls_cnt = 0
        for n in range(sample_num):
            data = gts[n]
            bbox, cam_param, joint, gt_hand_type, hand_type_valid = data['bbox'], data['cam_param'], data['joint'], data['hand_type'], data['hand_type_valid']
            focal = cam_param['focal']
            princpt = cam_param['princpt']
            gt_joint_coord = joint['cam_coord']
            joint_valid = joint['valid']
            
            # 获得预测数据
            pred_data = preds[n]

            # restore xy coordinates to original image space
            pred_joint_coord_img = pred_data['joint_coord'].copy()
            output_hm_shape = getattr(self, 'output_hm_shape', [])
            input_img_shape = getattr(self, 'input_img_shape', [])
            output_hm_shape = [(int)(a) for a in output_hm_shape]
            input_img_shape = [(int)(a) for a in input_img_shape]
            pred_joint_coord_img[:,0] = pred_joint_coord_img[:,0]/output_hm_shape[2]*input_img_shape[1]
            pred_joint_coord_img[:,1] = pred_joint_coord_img[:,1]/output_hm_shape[1]*input_img_shape[0]
            for j in range(self.joint_num*2):
                pred_joint_coord_img[j,:2] = trans_point2d(pred_joint_coord_img[j,:2],pred_data['inv_trans'])
            # restore depth to original camera space
            bbox_3d_size = getattr(self, 'bbox_3d_size', 0)
            pred_joint_coord_img[:,2] = (pred_joint_coord_img[:,2]/output_hm_shape[0] * 2 - 1) * (bbox_3d_size/2)

            # mrrpe
            output_root_hm_shape = getattr(self, 'output_root_hm_shape', 0)
            bbox_3d_size_root = getattr(self, 'bbox_3d_size_root', 0)
            if gt_hand_type == 'interacting' and joint_valid[self.root_joint_idx['left']] and joint_valid[self.root_joint_idx['right']]:
                pred_rel_root_depth = (pred_data['rel_root_depth']/output_root_hm_shape * 2 - 1) * (bbox_3d_size_root/2)

                pred_left_root_img = pred_joint_coord_img[self.root_joint_idx['left']].copy()
                pred_left_root_img[2] += data['abs_depth']['right'] + pred_rel_root_depth
                pred_left_root_cam = pixel2cam(pred_left_root_img[None,:], focal, princpt)[0]

                pred_right_root_img = pred_joint_coord_img[self.root_joint_idx['right']].copy()
                pred_right_root_img[2] += data['abs_depth']['right']
                pred_right_root_cam = pixel2cam(pred_right_root_img[None,:], focal, princpt)[0]
                
                pred_rel_root = pred_left_root_cam - pred_right_root_cam
                gt_rel_root = gt_joint_coord[self.root_joint_idx['left']] - gt_joint_coord[self.root_joint_idx['right']]
                mrrpe.append(float(np.sqrt(np.sum((pred_rel_root - gt_rel_root)**2))))

           
            # add root joint depth
            pred_joint_coord_img[self.joint_type['right'],2] += data['abs_depth']['right']
            pred_joint_coord_img[self.joint_type['left'],2] += data['abs_depth']['left']

            # back project to camera coordinate system
            pred_joint_coord_cam = pixel2cam(pred_joint_coord_img, focal, princpt)

            # root joint alignment
            for h in ('right', 'left'):
                pred_joint_coord_cam[self.joint_type[h]] = pred_joint_coord_cam[self.joint_type[h]] - pred_joint_coord_cam[self.root_joint_idx[h],None,:]
                gt_joint_coord[self.joint_type[h]] = gt_joint_coord[self.joint_type[h]] - gt_joint_coord[self.root_joint_idx[h],None,:]
            
            # mpjpe
            for j in range(self.joint_num*2):
                if joint_valid[j]:
                    if gt_hand_type == 'right' or gt_hand_type == 'left':
                        mpjpe_sh[j].append(np.sqrt(np.sum((pred_joint_coord_cam[j] - gt_joint_coord[j])**2)))
                    else:
                        mpjpe_ih[j].append(np.sqrt(np.sum((pred_joint_coord_cam[j] - gt_joint_coord[j])**2)))

            # handedness accuray
            if hand_type_valid:
                if gt_hand_type == 'right' and pred_data['hand_type'][0] > 0.5 and pred_data['hand_type'][1] < 0.5:
                    acc_hand_cls += 1
                elif gt_hand_type == 'left' and pred_data['hand_type'][0] < 0.5 and pred_data['hand_type'][1] > 0.5:
                    acc_hand_cls += 1
                elif gt_hand_type == 'interacting' and pred_data['hand_type'][0] > 0.5 and pred_data['hand_type'][1] > 0.5:
                    acc_hand_cls += 1
                hand_cls_cnt += 1

            vis = False
            if vis:
                img_path = data['img_path']
                cvimg = cv2.imread(img_path, cv2.IMREAD_COLOR | cv2.IMREAD_IGNORE_ORIENTATION)
                _img = cvimg[:,:,::-1].transpose(2,0,1)
                vis_kps = pred_joint_coord_img.copy()
                vis_valid = joint_valid.copy()
                capture = str(data['capture'])
                cam = str(data['cam'])
                frame = str(data['frame'])
                filename = 'out_' + str(n) + '_' + gt_hand_type + '.jpg'
                vis_keypoints(_img, vis_kps, vis_valid, self.skeleton, filename)

            vis = False
            if vis:
                filename = 'out_' + str(n) + '_3d.jpg'
                vis_3d_keypoints(pred_joint_coord_cam, joint_valid, self.skeleton, filename)
        

        if hand_cls_cnt > 0: print('Handedness accuracy: ' + str(acc_hand_cls / hand_cls_cnt))
        if len(mrrpe) > 0: print('MRRPE: ' + str(sum(mrrpe)/len(mrrpe)))
        print()

        tot_err = []
        eval_summary = 'MPJPE for each joint: \n'
        for j in range(self.joint_num*2):
            tot_err_j = np.mean(np.concatenate((np.stack(mpjpe_sh[j]), np.stack(mpjpe_ih[j]))))
            joint_name = self.skeleton[j]['name']
            eval_summary += (joint_name + ': %.2f, ' % tot_err_j)
            tot_err.append(tot_err_j)
        print(eval_summary)
        print('MPJPE for all hand sequences: %.2f' % (np.mean(tot_err)))
        print()
        dict_err = {}
        dict_err['MPJPE_ALL'] = (float)(np.mean(tot_err))

        eval_summary = 'MPJPE for each joint: \n'
        for j in range(self.joint_num*2):
            mpjpe_sh[j] = np.mean(np.stack(mpjpe_sh[j]))
            joint_name = self.skeleton[j]['name']
            eval_summary += (joint_name + ': %.2f, ' % mpjpe_sh[j])
        print(eval_summary)
        print('MPJPE for single hand sequences: %.2f' % (np.mean(mpjpe_sh)))
        print()
        dict_err['MPJPE_SINGLE'] = (float)(np.mean(mpjpe_sh))

        eval_summary = 'MPJPE for each joint: \n'
        for j in range(self.joint_num*2):
            mpjpe_ih[j] = np.mean(np.stack(mpjpe_ih[j]))
            joint_name = self.skeleton[j]['name']
            eval_summary += (joint_name + ': %.2f, ' % mpjpe_ih[j])
        print(eval_summary)
        print('MPJPE for interacting hand sequences: %.2f' % (np.mean(mpjpe_ih)))
        dict_err['MPJPE_INTERACT'] = (float)(np.mean(mpjpe_ih))
        return dict_err






# Copyright (c) OpenMMLab. All rights reserved.
import inspect

import torch.nn as nn

from antgo.framework.helper.utils import is_tuple_of
from antgo.framework.helper.utils.parrots_wrapper import SyncBatchNorm, _BatchNorm, _InstanceNorm
from .registry import NORM_LAYERS

NORM_LAYERS.register_module('BN', module=nn.BatchNorm2d)
NORM_LAYERS.register_module('BN1d', module=nn.BatchNorm1d)
NORM_LAYERS.register_module('BN2d', module=nn.BatchNorm2d)
NORM_LAYERS.register_module('BN3d', module=nn.BatchNorm3d)
NORM_LAYERS.register_module('SyncBN', module=SyncBatchNorm)
NORM_LAYERS.register_module('GN', module=nn.GroupNorm)
NORM_LAYERS.register_module('LN', module=nn.LayerNorm)
NORM_LAYERS.register_module('IN', module=nn.InstanceNorm2d)
NORM_LAYERS.register_module('IN1d', module=nn.InstanceNorm1d)
NORM_LAYERS.register_module('IN2d', module=nn.InstanceNorm2d)
NORM_LAYERS.register_module('IN3d', module=nn.InstanceNorm3d)


def infer_abbr(class_type):
    """Infer abbreviation from the class name.

    When we build a norm layer with `build_norm_layer()`, we want to preserve
    the norm type in variable names, e.g, self.bn1, self.gn. This method will
    infer the abbreviation to map class types to abbreviations.

    Rule 1: If the class has the property "_abbr_", return the property.
    Rule 2: If the parent class is _BatchNorm, GroupNorm, LayerNorm or
    InstanceNorm, the abbreviation of this layer will be "bn", "gn", "ln" and
    "in" respectively.
    Rule 3: If the class name contains "batch", "group", "layer" or "instance",
    the abbreviation of this layer will be "bn", "gn", "ln" and "in"
    respectively.
    Rule 4: Otherwise, the abbreviation falls back to "norm".

    Args:
        class_type (type): The norm layer type.

    Returns:
        str: The inferred abbreviation.
    """
    if not inspect.isclass(class_type):
        raise TypeError(
            f'class_type must be a type, but got {type(class_type)}')
    if hasattr(class_type, '_abbr_'):
        return class_type._abbr_
    if issubclass(class_type, _InstanceNorm):  # IN is a subclass of BN
        return 'in'
    elif issubclass(class_type, _BatchNorm):
        return 'bn'
    elif issubclass(class_type, nn.GroupNorm):
        return 'gn'
    elif issubclass(class_type, nn.LayerNorm):
        return 'ln'
    else:
        class_name = class_type.__name__.lower()
        if 'batch' in class_name:
            return 'bn'
        elif 'group' in class_name:
            return 'gn'
        elif 'layer' in class_name:
            return 'ln'
        elif 'instance' in class_name:
            return 'in'
        else:
            return 'norm_layer'


def build_norm_layer(cfg, num_features, postfix=''):
    """Build normalization layer.

    Args:
        cfg (dict): The norm layer config, which should contain:

            - type (str): Layer type.
            - layer args: Args needed to instantiate a norm layer.
            - requires_grad (bool, optional): Whether stop gradient updates.
        num_features (int): Number of input channels.
        postfix (int | str): The postfix to be appended into norm abbreviation
            to create named layer.

    Returns:
        tuple[str, nn.Module]: The first element is the layer name consisting
        of abbreviation and postfix, e.g., bn1, gn. The second element is the
        created norm layer.
    """
    if not isinstance(cfg, dict):
        raise TypeError('cfg must be a dict')
    if 'type' not in cfg:
        raise KeyError('the cfg dict must contain the key "type"')
    cfg_ = cfg.copy()

    layer_type = cfg_.pop('type')
    if layer_type not in NORM_LAYERS:
        raise KeyError(f'Unrecognized norm type {layer_type}')

    norm_layer = NORM_LAYERS.get(layer_type)
    abbr = infer_abbr(norm_layer)

    assert isinstance(postfix, (int, str))
    name = abbr + str(postfix)

    requires_grad = cfg_.pop('requires_grad', True)
    cfg_.setdefault('eps', 1e-5)
    if layer_type != 'GN':
        layer = norm_layer(num_features, **cfg_)
        if layer_type == 'SyncBN' and hasattr(layer, '_specify_ddp_gpu_num'):
            layer._specify_ddp_gpu_num(1)
    else:
        assert 'num_groups' in cfg_
        layer = norm_layer(num_channels=num_features, **cfg_)

    for param in layer.parameters():
        param.requires_grad = requires_grad

    return name, layer


def is_norm(layer, exclude=None):
    """Check if a layer is a normalization layer.

    Args:
        layer (nn.Module): The layer to be checked.
        exclude (type | tuple[type]): Types to be excluded.

    Returns:
        bool: Whether the layer is a norm layer.
    """
    if exclude is not None:
        if not isinstance(exclude, tuple):
            exclude = (exclude, )
        if not is_tuple_of(exclude, type):
            raise TypeError(
                f'"exclude" must be either None or type or a tuple of types, '
                f'but got {type(exclude)}: {exclude}')

    if exclude and isinstance(layer, exclude):
        return False

    all_norm_bases = (_BatchNorm, _InstanceNorm, nn.GroupNorm, nn.LayerNorm)
    return isinstance(layer, all_norm_bases)



# Copyright (c) OpenMMLab. All rights reserved.
from .builder import build_iou_calculator
from .iou2d_calculator import BboxOverlaps2D, bbox_overlaps

__all__ = ['build_iou_calculator', 'BboxOverlaps2D', 'bbox_overlaps']



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class FT(nn.Module):
	'''
	araphrasing Complex Network: Network Compression via Factor Transfer
	http://papers.nips.cc/paper/7541-paraphrasing-complex-network-network-compression-via-factor-transfer.pdf
	'''
	def __init__(self):
		super(FT, self).__init__()

	def forward(self, factor_s, factor_t):
		loss = F.l1_loss(self.normalize(factor_s), self.normalize(factor_t))

		return loss

	def normalize(self, factor):
		norm_factor = F.normalize(factor.view(factor.size(0),-1))

		return norm_factor


import copy
import platform
import random
import warnings
from functools import partial

import numpy as np
from numpy.random.mtrand import sample
import torch
from torch.utils.data.dataset import IterableDataset
from antgo.framework.helper.parallel import collate
from antgo.framework.helper.runner import get_dist_info
from antgo.framework.helper.utils import TORCH_VERSION, Registry, build_from_cfg, digit_version
from torch.utils.data import DataLoader


from .samplers import (ClassAwareSampler, DistributedGroupSampler,
                       DistributedSampler, GroupSampler, InfiniteBatchSampler,
                       InfiniteGroupBatchSampler, KVSampler, DistributedKVSampler)

# if platform.system() != 'Windows':
#     # https://github.com/pytorch/pytorch/issues/973
#     import resource
#     rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
#     base_soft_limit = rlimit[0]
#     hard_limit = rlimit[1]
#     soft_limit = min(max(4096, base_soft_limit), hard_limit)
#     resource.setrlimit(resource.RLIMIT_NOFILE, (soft_limit, hard_limit))

DATASETS = Registry('dataset')
PIPELINES = Registry('pipeline')


def _concat_dataset(cfg, default_args=None):
    from .dataset_wrappers import ConcatDataset
    ann_files = cfg['ann_file']
    img_prefixes = cfg.get('img_prefix', None)
    seg_prefixes = cfg.get('seg_prefix', None)
    proposal_files = cfg.get('proposal_file', None)
    separate_eval = cfg.get('separate_eval', True)

    datasets = []
    num_dset = len(ann_files)
    for i in range(num_dset):
        data_cfg = copy.deepcopy(cfg)
        # pop 'separate_eval' since it is not a valid key for common datasets.
        if 'separate_eval' in data_cfg:
            data_cfg.pop('separate_eval')
        data_cfg['ann_file'] = ann_files[i]
        if isinstance(img_prefixes, (list, tuple)):
            data_cfg['img_prefix'] = img_prefixes[i]
        if isinstance(seg_prefixes, (list, tuple)):
            data_cfg['seg_prefix'] = seg_prefixes[i]
        if isinstance(proposal_files, (list, tuple)):
            data_cfg['proposal_file'] = proposal_files[i]
        datasets.append(build_dataset(data_cfg, default_args))

    return ConcatDataset(datasets, separate_eval)


def build_dataset(cfg, default_args=None):
    from .dataset_wrappers import (ConcatDataset, RepeatDataset, IterConcatDataset)
    if isinstance(cfg, (list, tuple)):
        type_list = [cfg[i].type for i in range(len(cfg))]
        if 'TFDataset' in type_list:
            dataset = IterConcatDataset([build_dataset(c, default_args) for c in cfg])
        else:
            dataset = ConcatDataset([build_dataset(c, default_args) for c in cfg])
    elif cfg['type'] == 'ConcatDataset':
        dataset = ConcatDataset(
            [build_dataset(c, default_args) for c in cfg['datasets']])
    elif cfg['type'] == 'RepeatDataset':
        dataset = RepeatDataset(
            build_dataset(cfg['dataset'], default_args), cfg['times'])
    elif isinstance(cfg.get('ann_file'), (list, tuple)):
        dataset = _concat_dataset(cfg, default_args)
    else:
        dataset = build_from_cfg(cfg, DATASETS, default_args)

    return dataset


def build_dataloader(dataset,
                     samples_per_gpu,
                     workers_per_gpu,
                     num_gpus=1,
                     dist=True,
                     shuffle=True,
                     seed=None,
                     runner_type='EpochBasedRunner',
                     persistent_workers=False,
                     class_aware_sampler=None,
                     ignore_stack=[],
                     **kwargs):
    """Build PyTorch DataLoader.

    In distributed training, each GPU/process has a dataloader.
    In non-distributed training, there is only one dataloader for all GPUs.

    Args:
        dataset (Dataset): A PyTorch dataset.
        samples_per_gpu (int): Number of training samples on each GPU, i.e.,
            batch size of each GPU.
        workers_per_gpu (int): How many subprocesses to use for data loading
            for each GPU.
        num_gpus (int): Number of GPUs. Only used in non-distributed training.
        dist (bool): Distributed training/test or not. Default: True.
        shuffle (bool): Whether to shuffle the data at every epoch.
            Default: True.
        seed (int, Optional): Seed to be used. Default: None.
        runner_type (str): Type of runner. Default: `EpochBasedRunner`
        persistent_workers (bool): If True, the data loader will not shutdown
            the worker processes after a dataset has been consumed once.
            This allows to maintain the workers `Dataset` instances alive.
            This argument is only valid when PyTorch>=1.7.0. Default: False.
        class_aware_sampler (dict): Whether to use `ClassAwareSampler`
            during training. Default: None.
        kwargs: any keyword argument to be used to initialize DataLoader

    Returns:
        DataLoader: A PyTorch dataloader.
    """
    rank, world_size = get_dist_info()
    if dist:
        # When model is :obj:`DistributedDataParallel`,
        # `batch_size` of :obj:`dataloader` is the
        # number of training samples on each GPU.
        batch_size = samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu)
        num_workers = workers_per_gpu
    else:
        # When model is obj:`DataParallel`
        # the batch size is samples on all the GPUS
        batch_size = num_gpus * samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu)
        num_workers = num_gpus * workers_per_gpu

    strategy = None
    if isinstance(samples_per_gpu, list):
        strategy = {}
        for index, s in enumerate(samples_per_gpu):
            strategy[index] = s

    if runner_type == 'IterBasedRunner':
        # this is a batch sampler, which can yield
        # a mini-batch indices each time.
        # it can be used in both `DataParallel` and
        # `DistributedDataParallel`
        if shuffle:
            batch_sampler = InfiniteGroupBatchSampler(
                dataset, batch_size, world_size, rank, seed=seed)
        else:
            batch_sampler = InfiniteBatchSampler(
                dataset,
                batch_size,
                world_size,
                rank,
                seed=seed,
                shuffle=False)
        batch_size = 1
        sampler = None
    else:
        if class_aware_sampler is not None:
            # ClassAwareSampler can be used in both distributed and
            # non-distributed training.
            num_sample_class = class_aware_sampler.get('num_sample_class', 1)
            sampler = ClassAwareSampler(
                dataset,
                samples_per_gpu,
                world_size,
                rank,
                seed=seed,
                num_sample_class=num_sample_class)
        elif dist:
            # DistributedGroupSampler will definitely shuffle the data to
            # satisfy that images on each GPU are in the same group
            if shuffle:
                sampler = DistributedGroupSampler(
                    dataset, 
                    samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu),
                    world_size, 
                    rank, 
                    seed=seed, 
                    strategy=strategy)

            else:
                sampler = DistributedSampler(
                    dataset, world_size, rank, shuffle=False, seed=seed, strategy=strategy)
        else:
            sampler = GroupSampler(
                dataset,
                samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu),
                strategy=strategy) if shuffle else None

        batch_sampler = None

    if (TORCH_VERSION != 'parrots'
            and digit_version(TORCH_VERSION) >= digit_version('1.7.0')):
        kwargs['persistent_workers'] = persistent_workers
    elif persistent_workers is True:
        warnings.warn('persistent_workers is invalid because your pytorch '
                      'version is lower than 1.7.0')

    init_fn = partial(
        worker_init_fn, num_workers=num_workers, rank=rank,
        seed=seed) if seed is not None else None

    data_loader = DataLoader(
        dataset,
        batch_size=int(batch_size),
        sampler=sampler,
        num_workers=num_workers,
        batch_sampler=batch_sampler,
        collate_fn=partial(collate, samples_per_gpu=samples_per_gpu, ignore_stack=ignore_stack),
        pin_memory=kwargs.pop('pin_memory', False),
        worker_init_fn=init_fn,
        **kwargs)

    return data_loader


def build_kv_dataloader(dataset,
                     samples_per_gpu,
                     workers_per_gpu,
                     num_gpus=1,
                     dist=True,
                     shuffle=True,
                     seed=0,
                     runner_type='EpochBasedRunner',
                     persistent_workers=False,
                     ignore_stack=[],
                     **kwargs):
    sampler = None
    if dist:
        # When model is :obj:`DistributedDataParallel`,
        # `batch_size` of :obj:`dataloader` is the
        # number of training samples on each GPU.
        batch_size = samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu)
        num_workers = workers_per_gpu
    else:
        # When model is obj:`DataParallel`
        # the batch size is samples on all the GPUS
        batch_size = num_gpus * (samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu))
        num_workers = num_gpus * workers_per_gpu

    strategy = None
    if isinstance(samples_per_gpu, list):
        strategy = {}
        for index, s in enumerate(samples_per_gpu):
            strategy[index] = s
    
    rank, world_size = get_dist_info()
    if dist:
        sampler = DistributedKVSampler(
            dataset,
            batch_size, 
            num_replicas=world_size,
            shuffle=shuffle,
            rank=rank,
            seed=seed, 
            drop_last=kwargs.get('drop_last', False), 
            strategy=strategy)
    else:
        sampler = \
            KVSampler(
                dataset, 
                samples_per_gpu if not isinstance(samples_per_gpu, list) else np.sum(samples_per_gpu), 
                drop_last=kwargs.get('drop_last', False), 
                shuffle=shuffle,
                strategy=strategy)

    init_fn = partial(
        worker_init_fn, num_workers=num_workers, rank=rank,
        seed=seed) if seed is not None else None
        
    data_loader = DataLoader(
        dataset,
        batch_size=None,
        sampler=sampler,
        num_workers=num_workers,
        batch_sampler=None,
        pin_memory=kwargs.pop('pin_memory', False),
        collate_fn=partial(collate, samples_per_gpu=samples_per_gpu, ignore_stack=ignore_stack),
        worker_init_fn=init_fn)

    return data_loader


def build_iter_dataloader(dataset,   
                          samples_per_gpu,
                          workers_per_gpu,
                          ignore_stack=[],
                          **kwargs):
    batch_size = samples_per_gpu if not isinstance(samples_per_gpu, list) else int(np.sum(samples_per_gpu))
    dataset.samples_per_gpu = samples_per_gpu
    data_loader = torch.utils.data.DataLoader(
        dataset, 
        batch_size=batch_size, 
        num_workers=workers_per_gpu,
        pin_memory=kwargs.pop('pin_memory', False),
        collate_fn=partial(collate, samples_per_gpu=batch_size, ignore_stack=ignore_stack),
        drop_last=kwargs.get('drop_last', True))

    return data_loader


def worker_init_fn(worker_id, num_workers, rank, seed):
    # The seed of each worker equals to
    # num_worker * rank + worker_id + user_seed
    worker_seed = num_workers * rank + worker_id + seed
    np.random.seed(worker_seed)
    random.seed(worker_seed)
    torch.manual_seed(worker_seed)

    # 初始化
    worker_info = torch.utils.data.get_worker_info()
    if getattr(worker_info.dataset, 'worker_init_fn', None):
        worker_info.dataset.worker_init_fn(worker_id)